import pytest

from numpydoc.xref import DEFAULT_LINKS, make_xref

# Use the default numpydoc link mapping
xref_aliases = DEFAULT_LINKS


# Comes mainly from numpy
data = r"""
(...) array_like, float, optional
(...) :term:`numpy:array_like`, :class:`python:float`, optional

(2,) ndarray
(2,) :obj:`ndarray <numpy.ndarray>`

(...,M,N) array_like
(...,M,N) :term:`numpy:array_like`

(..., M, N) array_like
(..., :obj:`M`, :obj:`N`) :term:`numpy:array_like`

(float, float), optional
(:class:`python:float`, :class:`python:float`), optional

1-D array or sequence
1-D :obj:`array <numpy.ndarray>` or :term:`python:sequence`

array of str or unicode-like
:obj:`array <numpy.ndarray>` of :class:`python:str` or unicode-like

array_like of float
:term:`numpy:array_like` of :class:`python:float`

bool or callable
:ref:`bool <python:bltin-boolean-values>` or :func:`python:callable`

int in [0, 255]
:class:`python:int` in [0, 255]

int or None, optional
:class:`python:int` or :data:`python:None`, optional

list of str or array_like
:class:`python:list` of :class:`python:str` or :term:`numpy:array_like`

sequence of array_like
:term:`python:sequence` of :term:`numpy:array_like`

str or pathlib.Path
:class:`python:str` or :obj:`pathlib.Path`

{'', string}, optional
{'', :class:`python:str`}, optional

{'C', 'F', 'A', or 'K'}, optional
{'C', 'F', 'A', or 'K'}, optional

{'linear', 'lower', 'higher', 'midpoint', 'nearest'}
{'linear', 'lower', 'higher', 'midpoint', 'nearest'}

{False, True, 'greedy', 'optimal'}
{:data:`python:False`, :data:`python:True`, 'greedy', 'optimal'}

{{'begin', 1}, {'end', 0}}, {string, int}
{{'begin', 1}, {'end', 0}}, {:class:`python:str`, :class:`python:int`}

callable f'(x,*args)
:func:`python:callable` f'(x,*args)

callable ``fhess(x, *args)``, optional
:func:`python:callable` ``fhess(x, *args)``, optional

spmatrix (format: ``csr``, ``bsr``, ``dia`` or coo``)
:obj:`spmatrix` (format: ``csr``, ``bsr``, ``dia`` or coo``)

:ref:`strftime <strftime-strptime-behavior>`
:ref:`strftime <strftime-strptime-behavior>`

callable or :ref:`strftime <strftime-strptime-behavior>`
:func:`python:callable` or :ref:`strftime <strftime-strptime-behavior>`

callable or :ref:`strftime behavior <strftime-strptime-behavior>`
:func:`python:callable` or :ref:`strftime behavior <strftime-strptime-behavior>`

list(int)
:class:`python:list`\(:class:`python:int`)

list[int]
:class:`python:list`\[:class:`python:int`]

dict(str, int)
:class:`python:dict`\(:class:`python:str`, :class:`python:int`)

dict[str,  int]
:class:`python:dict`\[:class:`python:str`,  :class:`python:int`]

tuple(float, float)
:class:`python:tuple`\(:class:`python:float`, :class:`python:float`)

dict[tuple(str, str), int]
:class:`python:dict`\[:class:`python:tuple`\(:class:`python:str`, :class:`python:str`), :class:`python:int`]
"""

data_ignore_obj = r"""
(...) array_like, float, optional
(...) :term:`numpy:array_like`, :class:`python:float`, optional

(2,) ndarray
(2,) :obj:`ndarray <numpy.ndarray>`

(...,M,N) array_like
(...,M,N) :term:`numpy:array_like`

(..., M, N) array_like
(..., M, N) :term:`numpy:array_like`

(float, float), optional
(:class:`python:float`, :class:`python:float`), optional

1-D array or sequence
1-D :obj:`array <numpy.ndarray>` or :term:`python:sequence`

array of str or unicode-like
:obj:`array <numpy.ndarray>` of :class:`python:str` or unicode-like

array_like of float
:term:`numpy:array_like` of :class:`python:float`

bool or callable
:ref:`bool <python:bltin-boolean-values>` or :func:`python:callable`

int in [0, 255]
:class:`python:int` in [0, 255]

int or None, optional
:class:`python:int` or :data:`python:None`, optional

list of str or array_like
:class:`python:list` of :class:`python:str` or :term:`numpy:array_like`

sequence of array_like
:term:`python:sequence` of :term:`numpy:array_like`

str or pathlib.Path
:class:`python:str` or pathlib.Path

{'', string}, optional
{'', :class:`python:str`}, optional

{'C', 'F', 'A', or 'K'}, optional
{'C', 'F', 'A', or 'K'}, optional

{'linear', 'lower', 'higher', 'midpoint', 'nearest'}
{'linear', 'lower', 'higher', 'midpoint', 'nearest'}

{False, True, 'greedy', 'optimal'}
{:data:`python:False`, :data:`python:True`, 'greedy', 'optimal'}

{{'begin', 1}, {'end', 0}}, {string, int}
{{'begin', 1}, {'end', 0}}, {:class:`python:str`, :class:`python:int`}

callable f'(x,*args)
:func:`python:callable` f'(x,*args)

callable ``fhess(x, *args)``, optional
:func:`python:callable` ``fhess(x, *args)``, optional

spmatrix (format: ``csr``, ``bsr``, ``dia`` or coo``)
spmatrix (format: ``csr``, ``bsr``, ``dia`` or coo``)

:ref:`strftime <strftime-strptime-behavior>`
:ref:`strftime <strftime-strptime-behavior>`

callable or :ref:`strftime <strftime-strptime-behavior>`
:func:`python:callable` or :ref:`strftime <strftime-strptime-behavior>`

callable or :ref:`strftime behavior <strftime-strptime-behavior>`
:func:`python:callable` or :ref:`strftime behavior <strftime-strptime-behavior>`

list(int)
:class:`python:list`\(:class:`python:int`)

list[int]
:class:`python:list`\[:class:`python:int`]

dict(str, int)
:class:`python:dict`\(:class:`python:str`, :class:`python:int`)

dict[str,  int]
:class:`python:dict`\[:class:`python:str`,  :class:`python:int`]

tuple(float, float)
:class:`python:tuple`\(:class:`python:float`, :class:`python:float`)

dict[tuple(str, str), int]
:class:`python:dict`\[:class:`python:tuple`\(:class:`python:str`, :class:`python:str`), :class:`python:int`]
"""

xref_ignore = {"or", "in", "of", "default", "optional"}


@pytest.mark.parametrize(
    ("param_type", "expected_result"),
    [tuple(s.split("\n")) for s in data.strip().split("\n\n")],
)
def test_make_xref(param_type, expected_result):
    assert make_xref(param_type, xref_aliases, xref_ignore) == expected_result


@pytest.mark.parametrize(
    ("param_type", "expected_result"),
    [tuple(s.split("\n")) for s in data_ignore_obj.strip().split("\n\n")],
)
def test_make_xref_ignore_unknown(param_type, expected_result):
    assert make_xref(param_type, xref_aliases, xref_ignore="all") == expected_result


def test_xref_ignore_is_all():
    with pytest.raises(TypeError, match="must be a set or 'all'"):
        make_xref("array_like", xref_aliases, xref_ignore="foo")



"""Extract reference documentation from the NumPy source tree."""

import copy
import inspect
import pydoc
import re
import sys
import textwrap
from collections import namedtuple
from collections.abc import Callable, Mapping
from functools import cached_property
from warnings import warn


def strip_blank_lines(l):
    "Remove leading and trailing blank lines from a list of lines"
    while l and not l[0].strip():
        del l[0]
    while l and not l[-1].strip():
        del l[-1]
    return l


class Reader:
    """A line-based string reader."""

    def __init__(self, data):
        """
        Parameters
        ----------
        data : str
           String with lines separated by '\\n'.

        """
        if isinstance(data, list):
            self._str = data
        else:
            self._str = data.split("\n")  # store string as list of lines

        self.reset()

    def __getitem__(self, n):
        return self._str[n]

    def reset(self):
        self._l = 0  # current line nr

    def read(self):
        if not self.eof():
            out = self[self._l]
            self._l += 1
            return out
        else:
            return ""

    def seek_next_non_empty_line(self):
        for l in self[self._l :]:
            if l.strip():
                break
            else:
                self._l += 1

    def eof(self):
        return self._l >= len(self._str)

    def read_to_condition(self, condition_func):
        start = self._l
        for line in self[start:]:
            if condition_func(line):
                return self[start : self._l]
            self._l += 1
            if self.eof():
                return self[start : self._l + 1]
        return []

    def read_to_next_empty_line(self):
        self.seek_next_non_empty_line()

        def is_empty(line):
            return not line.strip()

        return self.read_to_condition(is_empty)

    def read_to_next_unindented_line(self):
        def is_unindented(line):
            return line.strip() and (len(line.lstrip()) == len(line))

        return self.read_to_condition(is_unindented)

    def peek(self, n=0):
        if self._l + n < len(self._str):
            return self[self._l + n]
        else:
            return ""

    def is_empty(self):
        return not "".join(self._str).strip()


class ParseError(Exception):
    def __str__(self):
        message = self.args[0]
        if hasattr(self, "docstring"):
            message = f"{message} in {self.docstring!r}"
        return message


Parameter = namedtuple("Parameter", ["name", "type", "desc"])


class NumpyDocString(Mapping):
    """Parses a numpydoc string to an abstract representation

    Instances define a mapping from section title to structured data.

    """

    sections = {
        "Signature": "",
        "Summary": [""],
        "Extended Summary": [],
        "Parameters": [],
        "Attributes": [],
        "Methods": [],
        "Returns": [],
        "Yields": [],
        "Receives": [],
        "Other Parameters": [],
        "Raises": [],
        "Warns": [],
        "Warnings": [],
        "See Also": [],
        "Notes": [],
        "References": "",
        "Examples": "",
        "index": {},
    }

    def __init__(self, docstring, config=None):
        orig_docstring = docstring
        docstring = textwrap.dedent(docstring).split("\n")

        self._doc = Reader(docstring)
        self._parsed_data = copy.deepcopy(self.sections)

        try:
            self._parse()
        except ParseError as e:
            e.docstring = orig_docstring
            raise

    def __getitem__(self, key):
        return self._parsed_data[key]

    def __setitem__(self, key, val):
        if key not in self._parsed_data:
            self._error_location(f"Unknown section {key}", error=False)
        else:
            self._parsed_data[key] = val

    def __iter__(self):
        return iter(self._parsed_data)

    def __len__(self):
        return len(self._parsed_data)

    def _is_at_section(self):
        self._doc.seek_next_non_empty_line()

        if self._doc.eof():
            return False

        l1 = self._doc.peek().strip()  # e.g. Parameters

        if l1.startswith(".. index::"):
            return True

        l2 = self._doc.peek(1).strip()  # ---------- or ==========
        if len(l2) >= 3 and (set(l2) in ({"-"}, {"="})) and len(l2) != len(l1):
            snip = "\n".join(self._doc._str[:2]) + "..."
            self._error_location(
                f"potentially wrong underline length... \n{l1} \n{l2} in \n{snip}",
                error=False,
            )
        return l2.startswith("-" * len(l1)) or l2.startswith("=" * len(l1))

    def _strip(self, doc):
        i = 0
        j = 0
        for i, line in enumerate(doc):
            if line.strip():
                break

        for j, line in enumerate(doc[::-1]):
            if line.strip():
                break

        return doc[i : len(doc) - j]

    def _read_to_next_section(self):
        section = self._doc.read_to_next_empty_line()

        while not self._is_at_section() and not self._doc.eof():
            if not self._doc.peek(-1).strip():  # previous line was empty
                section += [""]

            section += self._doc.read_to_next_empty_line()

        return section

    def _read_sections(self):
        while not self._doc.eof():
            data = self._read_to_next_section()
            name = data[0].strip()

            if name.startswith(".."):  # index section
                yield name, data[1:]
            elif len(data) < 2:
                yield StopIteration
            else:
                yield name, self._strip(data[2:])

    def _parse_param_list(self, content, single_element_is_type=False):
        content = dedent_lines(content)
        r = Reader(content)
        params = []
        while not r.eof():
            header = r.read().strip()
            if " : " in header:
                arg_name, arg_type = header.split(" : ", maxsplit=1)
            else:
                # NOTE: param line with single element should never have a
                # a " :" before the description line, so this should probably
                # warn.
                header = header.removesuffix(" :")
                if single_element_is_type:
                    arg_name, arg_type = "", header
                else:
                    arg_name, arg_type = header, ""

            desc = r.read_to_next_unindented_line()
            desc = dedent_lines(desc)
            desc = strip_blank_lines(desc)

            params.append(Parameter(arg_name, arg_type, desc))

        return params

    # See also supports the following formats.
    #
    # <FUNCNAME>
    # <FUNCNAME> SPACE* COLON SPACE+ <DESC> SPACE*
    # <FUNCNAME> ( COMMA SPACE+ <FUNCNAME>)+ (COMMA | PERIOD)? SPACE*
    # <FUNCNAME> ( COMMA SPACE+ <FUNCNAME>)* SPACE* COLON SPACE+ <DESC> SPACE*

    # <FUNCNAME> is one of
    #   <PLAIN_FUNCNAME>
    #   COLON <ROLE> COLON BACKTICK <PLAIN_FUNCNAME> BACKTICK
    # where
    #   <PLAIN_FUNCNAME> is a legal function name, and
    #   <ROLE> is any nonempty sequence of word characters.
    # Examples: func_f1  :meth:`func_h1` :obj:`~baz.obj_r` :class:`class_j`
    # <DESC> is a string describing the function.

    _role = r":(?P<role>(py:)?\w+):"
    _funcbacktick = r"`(?P<name>(?:~\w+\.)?[a-zA-Z0-9_\.-]+)`"
    _funcplain = r"(?P<name2>[a-zA-Z0-9_\.-]+)"
    _funcname = r"(" + _role + _funcbacktick + r"|" + _funcplain + r")"
    _funcnamenext = _funcname.replace("role", "rolenext")
    _funcnamenext = _funcnamenext.replace("name", "namenext")
    _description = r"(?P<description>\s*:(\s+(?P<desc>\S+.*))?)?\s*$"
    _func_rgx = re.compile(r"^\s*" + _funcname + r"\s*")
    _line_rgx = re.compile(
        r"^\s*"
        + r"(?P<allfuncs>"
        + _funcname  # group for all function names
        + r"(?P<morefuncs>([,]\s+"
        + _funcnamenext
        + r")*)"
        + r")"
        + r"(?P<trailing>[,\.])?"  # end of "allfuncs"
        + _description  # Some function lists have a trailing comma (or period)  '\s*'
    )

    # Empty <DESC> elements are replaced with '..'
    empty_description = ".."

    def _parse_see_also(self, content):
        """
        func_name : Descriptive text
            continued text
        another_func_name : Descriptive text
        func_name1, func_name2, :meth:`func_name`, func_name3

        """

        content = dedent_lines(content)

        items = []

        def parse_item_name(text):
            """Match ':role:`name`' or 'name'."""
            m = self._func_rgx.match(text)
            if not m:
                self._error_location(f"Error parsing See Also entry {line!r}")
            role = m.group("role")
            name = m.group("name") if role else m.group("name2")
            return name, role, m.end()

        rest = []
        for line in content:
            if not line.strip():
                continue

            line_match = self._line_rgx.match(line)
            description = None
            if line_match:
                description = line_match.group("desc")
                if line_match.group("trailing") and description:
                    self._error_location(
                        "Unexpected comma or period after function list at index %d of "
                        'line "%s"' % (line_match.end("trailing"), line),
                        error=False,
                    )
            if not description and line.startswith(" "):
                rest.append(line.strip())
            elif line_match:
                funcs = []
                text = line_match.group("allfuncs")
                while True:
                    if not text.strip():
                        break
                    name, role, match_end = parse_item_name(text)
                    funcs.append((name, role))
                    text = text[match_end:].strip()
                    if text and text[0] == ",":
                        text = text[1:].strip()
                rest = list(filter(None, [description]))
                items.append((funcs, rest))
            else:
                self._error_location(f"Error parsing See Also entry {line!r}")
        return items

    def _parse_index(self, section, content):
        """
        .. index:: default
           :refguide: something, else, and more

        """

        def strip_each_in(lst):
            return [s.strip() for s in lst]

        out = {}
        section = section.split("::")
        if len(section) > 1:
            out["default"] = strip_each_in(section[1].split(","))[0]
        for line in content:
            line = line.split(":")
            if len(line) > 2:
                out[line[1]] = strip_each_in(line[2].split(","))
        return out

    def _parse_summary(self):
        """Grab signature (if given) and summary"""
        if self._is_at_section():
            return

        # If several signatures present, take the last one
        while True:
            summary = self._doc.read_to_next_empty_line()
            summary_str = " ".join([s.strip() for s in summary]).strip()
            compiled = re.compile(r"^([\w., ]+=)?\s*[\w\.]+\(.*\)$")
            if compiled.match(summary_str):
                self["Signature"] = summary_str
                if not self._is_at_section():
                    continue
            break

        if summary is not None:
            self["Summary"] = summary

        if not self._is_at_section():
            self["Extended Summary"] = self._read_to_next_section()

    def _parse(self):
        self._doc.reset()
        self._parse_summary()

        sections = list(self._read_sections())
        section_names = {section for section, content in sections}

        has_yields = "Yields" in section_names
        # We could do more tests, but we are not. Arbitrarily.
        if not has_yields and "Receives" in section_names:
            msg = "Docstring contains a Receives section but not Yields."
            raise ValueError(msg)

        for section, content in sections:
            if not section.startswith(".."):
                section = (s.capitalize() for s in section.split(" "))
                section = " ".join(section)
                if self.get(section):
                    self._error_location(
                        "The section %s appears twice in  %s"
                        % (section, "\n".join(self._doc._str))
                    )

            if section in ("Parameters", "Other Parameters", "Attributes", "Methods"):
                self[section] = self._parse_param_list(content)
            elif section in ("Returns", "Yields", "Raises", "Warns", "Receives"):
                self[section] = self._parse_param_list(
                    content, single_element_is_type=True
                )
            elif section.startswith(".. index::"):
                self["index"] = self._parse_index(section, content)
            elif section == "See Also":
                self["See Also"] = self._parse_see_also(content)
            else:
                self[section] = content

    @property
    def _obj(self):
        if hasattr(self, "_cls"):
            return self._cls
        elif hasattr(self, "_f"):
            return self._f
        return None

    def _error_location(self, msg, error=True):
        if self._obj is not None:
            # we know where the docs came from:
            try:
                filename = inspect.getsourcefile(self._obj)
            except TypeError:
                filename = None
            # Make UserWarning more descriptive via object introspection.
            # Skip if introspection fails
            name = getattr(self._obj, "__name__", None)
            if name is None:
                name = getattr(getattr(self._obj, "__class__", None), "__name__", None)
            if name is not None:
                msg += f" in the docstring of {name}"
            msg += f" in {filename}." if filename else ""
        if error:
            raise ValueError(msg)
        else:
            warn(msg, stacklevel=3)

    # string conversion routines

    def _str_header(self, name, symbol="-"):
        return [name, len(name) * symbol]

    def _str_indent(self, doc, indent=4):
        return [" " * indent + line for line in doc]

    def _str_signature(self):
        if self["Signature"]:
            return [self["Signature"].replace("*", r"\*")] + [""]
        return [""]

    def _str_summary(self):
        if self["Summary"]:
            return self["Summary"] + [""]
        return []

    def _str_extended_summary(self):
        if self["Extended Summary"]:
            return self["Extended Summary"] + [""]
        return []

    def _str_param_list(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            for param in self[name]:
                parts = []
                if param.name:
                    parts.append(param.name)
                if param.type:
                    parts.append(param.type)
                out += [" : ".join(parts)]
                if param.desc and "".join(param.desc).strip():
                    out += self._str_indent(param.desc)
            out += [""]
        return out

    def _str_section(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            out += self[name]
            out += [""]
        return out

    def _str_see_also(self, func_role):
        if not self["See Also"]:
            return []
        out = []
        out += self._str_header("See Also")
        out += [""]
        last_had_desc = True
        for funcs, desc in self["See Also"]:
            assert isinstance(funcs, list)
            links = []
            for func, role in funcs:
                if role:
                    link = f":{role}:`{func}`"
                elif func_role:
                    link = f":{func_role}:`{func}`"
                else:
                    link = f"`{func}`_"
                links.append(link)
            link = ", ".join(links)
            out += [link]
            if desc:
                out += self._str_indent([" ".join(desc)])
                last_had_desc = True
            else:
                last_had_desc = False
                out += self._str_indent([self.empty_description])

        if last_had_desc:
            out += [""]
        out += [""]
        return out

    def _str_index(self):
        idx = self["index"]
        out = []
        output_index = False
        default_index = idx.get("default", "")
        if default_index:
            output_index = True
        out += [f".. index:: {default_index}"]
        for section, references in idx.items():
            if section == "default":
                continue
            output_index = True
            out += [f"   :{section}: {', '.join(references)}"]
        if output_index:
            return out
        return ""

    def __str__(self, func_role=""):
        out = []
        out += self._str_signature()
        out += self._str_summary()
        out += self._str_extended_summary()
        out += self._str_param_list("Parameters")
        for param_list in ("Attributes", "Methods"):
            out += self._str_param_list(param_list)
        for param_list in (
            "Returns",
            "Yields",
            "Receives",
            "Other Parameters",
            "Raises",
            "Warns",
        ):
            out += self._str_param_list(param_list)
        out += self._str_section("Warnings")
        out += self._str_see_also(func_role)
        for s in ("Notes", "References", "Examples"):
            out += self._str_section(s)
        out += self._str_index()
        return "\n".join(out)


def dedent_lines(lines):
    """Deindent a list of lines maximally"""
    return textwrap.dedent("\n".join(lines)).split("\n")


class FunctionDoc(NumpyDocString):
    def __init__(self, func, role="func", doc=None, config=None):
        self._f = func
        self._role = role  # e.g. "func" or "meth"

        if doc is None:
            if func is None:
                raise ValueError("No function or docstring given")
            doc = inspect.getdoc(func) or ""
        if config is None:
            config = {}
        NumpyDocString.__init__(self, doc, config)

    def get_func(self):
        func_name = getattr(self._f, "__name__", self.__class__.__name__)
        if inspect.isclass(self._f):
            func = getattr(self._f, "__call__", self._f.__init__)
        else:
            func = self._f
        return func, func_name

    def __str__(self):
        out = ""

        _func, func_name = self.get_func()

        roles = {"func": "function", "meth": "method"}

        if self._role:
            if self._role not in roles:
                print(f"Warning: invalid role {self._role}")
            out += f".. {roles.get(self._role, '')}:: {func_name}\n    \n\n"

        out += super().__str__(func_role=self._role)
        return out


class ObjDoc(NumpyDocString):
    def __init__(self, obj, doc=None, config=None):
        self._f = obj
        if config is None:
            config = {}
        NumpyDocString.__init__(self, doc, config=config)


class ClassDoc(NumpyDocString):
    extra_public_methods = ["__call__"]

    def __init__(self, cls, doc=None, modulename="", func_doc=FunctionDoc, config=None):
        if not inspect.isclass(cls) and cls is not None:
            raise ValueError(f"Expected a class or None, but got {cls!r}")
        self._cls = cls

        if "sphinx" in sys.modules:
            from sphinx.ext.autodoc import ALL
        else:
            ALL = object()

        if config is None:
            config = {}
        self.show_inherited_members = config.get("show_inherited_class_members", True)

        if modulename and not modulename.endswith("."):
            modulename += "."
        self._mod = modulename

        if doc is None:
            if cls is None:
                raise ValueError("No class or documentation string given")
            doc = pydoc.getdoc(cls)

        NumpyDocString.__init__(self, doc)

        _members = config.get("members", [])
        if _members is ALL:
            _members = None
        _exclude = config.get("exclude-members", [])

        if config.get("show_class_members", True) and _exclude is not ALL:

            def splitlines_x(s):
                if not s:
                    return []
                else:
                    return s.splitlines()

            for field, items in [
                ("Methods", self.methods),
                ("Attributes", self.properties),
            ]:
                if not self[field]:
                    doc_list = []
                    for name in sorted(items):
                        if name in _exclude or (_members and name not in _members):
                            continue
                        try:
                            doc_item = pydoc.getdoc(getattr(self._cls, name))
                            doc_list.append(Parameter(name, "", splitlines_x(doc_item)))
                        except AttributeError:
                            pass  # method doesn't exist
                    self[field] = doc_list

    @property
    def methods(self):
        if self._cls is None:
            return []
        return [
            name
            for name, func in inspect.getmembers(self._cls)
            if (
                (not name.startswith("_") or name in self.extra_public_methods)
                and isinstance(func, Callable)
                and self._is_show_member(name)
            )
        ]

    @property
    def properties(self):
        if self._cls is None:
            return []
        return [
            name
            for name, func in inspect.getmembers(self._cls)
            if (
                not name.startswith("_")
                and not self._should_skip_member(name, self._cls)
                and (
                    func is None
                    or isinstance(func, property | cached_property)
                    or inspect.isdatadescriptor(func)
                )
                and self._is_show_member(name)
            )
        ]

    @staticmethod
    def _should_skip_member(name, klass):
        return (
            # Namedtuples should skip everything in their ._fields as the
            # docstrings for each of the members is: "Alias for field number X"
            issubclass(klass, tuple)
            and hasattr(klass, "_asdict")
            and hasattr(klass, "_fields")
            and name in klass._fields
        )

    def _is_show_member(self, name):
        return (
            # show all class members
            self.show_inherited_members
            # or class member is not inherited
            or name in self._cls.__dict__
        )


def get_doc_object(
    obj,
    what=None,
    doc=None,
    config=None,
    class_doc=ClassDoc,
    func_doc=FunctionDoc,
    obj_doc=ObjDoc,
):
    if what is None:
        if inspect.isclass(obj):
            what = "class"
        elif inspect.ismodule(obj):
            what = "module"
        elif isinstance(obj, Callable):
            what = "function"
        else:
            what = "object"
    if config is None:
        config = {}

    if what == "class":
        return class_doc(obj, func_doc=func_doc, doc=doc, config=config)
    elif what in ("function", "method"):
        return func_doc(obj, doc=doc, config=config)
    else:
        if doc is None:
            doc = pydoc.getdoc(obj)
        return obj_doc(obj, doc, config=config)



#!/usr/bin/env python
"""
Analyze docstrings to detect errors.

Call ``validate(object_name_to_validate)`` to get a dictionary
with all the detected errors.
"""

import ast
import collections
import functools
import importlib
import inspect
import os
import pydoc
import re
import textwrap
import tokenize
from copy import deepcopy
from typing import Any, Dict, List, Set

from .docscrape import get_doc_object

DIRECTIVES = ["versionadded", "versionchanged", "deprecated"]
DIRECTIVE_PATTERN = re.compile(
    r"^\s*\.\. ({})(?!::)".format("|".join(DIRECTIVES)), re.IGNORECASE | re.MULTILINE
)
ALLOWED_SECTIONS = [
    "Parameters",
    "Attributes",
    "Methods",
    "Returns",
    "Yields",
    "Other Parameters",
    "Raises",
    "Warns",
    "Warnings",
    "See Also",
    "Notes",
    "References",
    "Examples",
]
# NOTE: The following comment is a sentinel for embedding in the docs - do not
# modify/remove
# start-err-msg
ERROR_MSGS = {
    "GL01": "Docstring text (summary) should start in the line immediately "
    "after the opening quotes (not in the same line, or leaving a "
    "blank line in between)",
    "GL02": "Closing quotes should be placed in the line after the last text "
    "in the docstring (do not close the quotes in the same line as "
    "the text, or leave a blank line between the last text and the "
    "quotes)",
    "GL03": "Double line break found; please use only one blank line to "
    "separate sections or paragraphs, and do not leave blank lines "
    "at the end of docstrings",
    "GL05": 'Tabs found at the start of line "{line_with_tabs}", please use '
    "whitespace only",
    "GL06": 'Found unknown section "{section}". Allowed sections are: '
    "{allowed_sections}",
    "GL07": "Sections are in the wrong order. Correct order is: {correct_sections}",
    "GL08": "The object does not have a docstring",
    "GL09": "Deprecation warning should precede extended summary",
    "GL10": "reST directives {directives} must be followed by two colons",
    "SS01": "No summary found (a short summary in a single line should be "
    "present at the beginning of the docstring)",
    "SS02": "Summary does not start with a capital letter",
    "SS03": "Summary does not end with a period",
    "SS04": "Summary contains heading whitespaces",
    "SS05": "Summary must start with infinitive verb, not third person "
    '(e.g. use "Generate" instead of "Generates")',
    "SS06": "Summary should fit in a single line",
    "ES01": "No extended summary found",
    "PR01": "Parameters {missing_params} not documented",
    "PR02": "Unknown parameters {unknown_params}",
    "PR03": "Wrong parameters order. Actual: {actual_params}. "
    "Documented: {documented_params}",
    "PR04": 'Parameter "{param_name}" has no type',
    "PR05": 'Parameter "{param_name}" type should not finish with "."',
    "PR06": 'Parameter "{param_name}" type should use "{right_type}" instead '
    'of "{wrong_type}"',
    "PR07": 'Parameter "{param_name}" has no description',
    "PR08": 'Parameter "{param_name}" description should start with a capital letter',
    "PR09": 'Parameter "{param_name}" description should finish with "."',
    "PR10": 'Parameter "{param_name}" requires a space before the colon '
    "separating the parameter name and type",
    "RT01": "No Returns section found",
    "RT02": "The first line of the Returns section should contain only the "
    "type, unless multiple values are being returned",
    "RT03": "Return value has no description",
    "RT04": "Return value description should start with a capital letter",
    "RT05": 'Return value description should finish with "."',
    "YD01": "No Yields section found",
    "SA01": "See Also section not found",
    "SA02": "Missing period at end of description for See Also "
    '"{reference_name}" reference',
    "SA03": "Description should be capitalized for See Also "
    '"{reference_name}" reference',
    "SA04": 'Missing description for See Also "{reference_name}" reference',
    "EX01": "No examples section found",
}
# end-err-msg
# NOTE: The above comment is a sentinel for embedding in the docs - do not
# modify/remove

# Ignore these when evaluating end-of-line-"." checks
IGNORE_STARTS = (" ", "* ", "- ")

# inline comments that can suppress individual checks per line
IGNORE_COMMENT_PATTERN = re.compile("(?:.* numpydoc ignore[=|:] ?)(.+)")


def _unwrap(obj):
    """Iteratively traverse obj.__wrapped__ until first non-wrapped obj found."""
    while hasattr(obj, "__wrapped__"):
        obj = obj.__wrapped__
    return obj


# This function gets called once per function/method to be validated.
# We have to balance memory usage with performance here. It shouldn't be too
# bad to store these `dict`s (they should be rare), but to be safe let's keep
# the limit low-ish. This was set by looking at scipy, numpy, matplotlib,
# and pandas, and they had between ~500 and ~1300 .py files as of 2023-08-16.
@functools.lru_cache(maxsize=2000)
def extract_ignore_validation_comments(
    filepath: os.PathLike | None,
    encoding: str = "utf-8",
) -> Dict[int, List[str]]:
    """
    Extract inline comments indicating certain validation checks should be ignored.

    Parameters
    ----------
    filepath : os.PathLike
        Path to the file being inspected.

    Returns
    -------
    dict[int, list[str]]
        Mapping of line number to a list of checks to ignore.
    """
    numpydoc_ignore_comments = {}
    try:
        file = open(filepath, encoding=encoding)
    except (OSError, TypeError):  # can be None, nonexistent, or unreadable
        return numpydoc_ignore_comments
    with file:
        last_declaration = 1
        declarations = ["def", "class"]
        for token in tokenize.generate_tokens(file.readline):
            if token.type == tokenize.NAME and token.string in declarations:
                last_declaration = token.start[0]
            if token.type == tokenize.COMMENT:
                match = re.match(IGNORE_COMMENT_PATTERN, token.string)
                if match:
                    rules = match.group(1).split(",")
                    numpydoc_ignore_comments[last_declaration] = rules
    return numpydoc_ignore_comments


def get_validation_checks(validation_checks: Set[str]) -> Set[str]:
    """
    Get the set of validation checks to report on.

    Parameters
    ----------
    validation_checks : set[str]
        A set of validation checks to report on. If the set is ``{"all"}``,
        all checks will be reported. If the set contains just specific checks,
        only those will be reported on. If the set contains both ``"all"`` and
        specific checks, all checks except those included in the set will be
        reported on.

    Returns
    -------
    set[str]
        The set of validation checks to report on.
    """
    valid_error_codes = set(ERROR_MSGS.keys())
    if "all" in validation_checks:
        block = deepcopy(validation_checks)
        validation_checks = valid_error_codes - block

    # Ensure that the validation check set contains only valid error codes
    invalid_error_codes = validation_checks - valid_error_codes
    if invalid_error_codes:
        raise ValueError(
            f"Unrecognized validation code(s) in numpydoc_validation_checks "
            f"config value: {invalid_error_codes}"
        )

    return validation_checks


def error(code, **kwargs):
    """
    Return a tuple with the error code and the message with variables replaced.

    This is syntactic sugar so instead of:
    - `('PR02', ERROR_MSGS['PR02'].format(doctest_log=log))`

    We can simply use:
    - `error('PR02', doctest_log=log)`

    Parameters
    ----------
    code : str
        Error code.
    **kwargs
        Values for the variables in the error messages

    Returns
    -------
    code : str
        Error code.
    message : str
        Error message with variables replaced.
    """
    return code, ERROR_MSGS[code].format(**kwargs)


class Validator:
    # TODO Can all this class be merged into NumpyDocString?
    def __init__(self, doc_object):
        self.doc = doc_object
        self.obj = self.doc._obj
        self.code_obj = inspect.unwrap(self.obj)
        self.raw_doc = self.obj.__doc__ or ""
        self.clean_doc = pydoc.getdoc(self.obj)

    @property
    def name(self):
        return ".".join([self.obj.__module__, self.obj.__name__])

    @staticmethod
    def _load_obj(name):
        """
        Import Python object from its name as string.

        Parameters
        ----------
        name : str
            Object name to import (e.g. pandas.Series.str.upper)

        Returns
        -------
        object
            Python object that can be a class, method, function...

        Examples
        --------
        >>> Validator._load_obj("datetime.datetime")
        <class 'datetime.datetime'>
        """
        for maxsplit in range(name.count(".") + 1):
            module, *func_parts = name.rsplit(".", maxsplit)
            try:
                obj = importlib.import_module(module)
            except ImportError:
                pass
            else:
                break
        else:
            raise ImportError(f'No module can be imported from "{name}"')

        for part in func_parts:
            obj = getattr(obj, part)
        return obj

    @property
    def type(self):
        return type(self.obj).__name__

    @property
    def is_function_or_method(self):
        return inspect.isfunction(self.obj)

    @property
    def is_mod(self):
        return inspect.ismodule(self.obj)

    @property
    def is_generator_function(self):
        return inspect.isgeneratorfunction(_unwrap(self.obj))

    @property
    def source_file_name(self):
        """
        File name where the object is implemented (e.g. pandas/core/frame.py).
        """
        try:
            if isinstance(self.code_obj, property):
                fname = inspect.getsourcefile(self.code_obj.fget)
            elif isinstance(self.code_obj, functools.cached_property):
                fname = inspect.getsourcefile(self.code_obj.func)
            else:
                fname = inspect.getsourcefile(self.code_obj)

        except TypeError:
            # In some cases the object is something complex like a cython
            # object that can't be easily introspected. And it's better to
            # return the source code file of the object as None, than crash
            pass
        else:
            return fname

    # When calling validate, files are parsed twice
    @staticmethod
    @functools.lru_cache(maxsize=4000)
    def _getsourcelines(obj: Any):
        return inspect.getsourcelines(obj)

    @property
    def source_file_def_line(self):
        """
        Number of line where the object is defined in its file.
        """
        try:
            if isinstance(self.code_obj, property):
                sourcelines = self._getsourcelines(self.code_obj.fget)
            elif isinstance(self.code_obj, functools.cached_property):
                sourcelines = self._getsourcelines(self.code_obj.func)
            else:
                sourcelines = self._getsourcelines(self.code_obj)
            # getsourcelines will return the line of the first decorator found for the
            # current function. We have to find the def declaration after that.
            def_line = next(
                i
                for i, x in enumerate(
                    re.match("^ *(def|class) ", s) for s in sourcelines[0]
                )
                if x is not None
            )
            return sourcelines[-1] + def_line
        except (OSError, TypeError):
            # In some cases the object is something complex like a cython
            # object that can't be easily introspected. And it's better to
            # return the line number as None, than crash
            pass

    @property
    def start_blank_lines(self):
        i = None
        if self.raw_doc:
            for i, row in enumerate(self.raw_doc.split("\n")):
                if row.strip():
                    break
        return i

    @property
    def end_blank_lines(self):
        i = None
        if self.raw_doc:
            for i, row in enumerate(reversed(self.raw_doc.split("\n"))):
                if row.strip():
                    break
        return i

    @property
    def double_blank_lines(self):
        prev = True
        for row in self.raw_doc.split("\n"):
            if not prev and not row.strip():
                return True
            prev = row.strip()
        return False

    @property
    def section_titles(self):
        sections = []
        self.doc._doc.reset()
        while not self.doc._doc.eof():
            content = self.doc._read_to_next_section()
            if (
                len(content) > 1
                and len(content[0]) == len(content[1])
                and set(content[1]) == {"-"}
            ):
                sections.append(content[0])
        return sections

    @property
    def summary(self):
        return " ".join(self.doc["Summary"])

    @property
    def num_summary_lines(self):
        return len(self.doc["Summary"])

    @property
    def extended_summary(self):
        if not self.doc["Extended Summary"] and len(self.doc["Summary"]) > 1:
            return " ".join(self.doc["Summary"])
        return " ".join(self.doc["Extended Summary"])

    def _doc_parameters(self, sections):
        parameters = collections.OrderedDict()
        for section in sections:
            for names, type_, desc in self.doc[section]:
                for name in names.split(", "):
                    parameters[name] = (type_, desc)
        return parameters

    @property
    def doc_parameters(self):
        return self._doc_parameters(["Parameters"])

    @property
    def doc_other_parameters(self):
        return self._doc_parameters(["Other Parameters"])

    @property
    def doc_all_parameters(self):
        return self._doc_parameters(["Parameters", "Other Parameters"])

    @property
    def signature_parameters(self):
        def add_stars(param_name, info):
            """
            Add stars to *args and **kwargs parameters
            """
            if info.kind == inspect.Parameter.VAR_POSITIONAL:
                return f"*{param_name}"
            elif info.kind == inspect.Parameter.VAR_KEYWORD:
                return f"**{param_name}"
            else:
                return param_name

        if inspect.isclass(self.obj):
            if hasattr(self.obj, "_accessors") and (
                self.name.split(".")[-1] in self.obj._accessors
            ):
                # accessor classes have a signature but don't want to show this
                return tuple()
        try:
            sig = inspect.signature(self.obj)
        except (TypeError, ValueError):
            # Some objects, mainly in C extensions do not support introspection
            # of the signature
            return tuple()

        params = tuple(
            add_stars(parameter, sig.parameters[parameter])
            for parameter in sig.parameters
        )
        if params and params[0] in ("self", "cls"):
            return params[1:]
        return params

    @property
    def parameter_mismatches(self):
        errs = []
        signature_params = self.signature_parameters
        all_params = tuple(param.replace("\\", "") for param in self.doc_all_parameters)
        missing = set(signature_params) - set(all_params)
        if missing:
            errs.append(error("PR01", missing_params=str(missing)))
        extra = set(all_params) - set(signature_params)
        if extra:
            errs.append(error("PR02", unknown_params=str(extra)))
        if (
            not missing
            and not extra
            and signature_params != all_params
            and not (not signature_params and not all_params)
        ):
            errs.append(
                error(
                    "PR03", actual_params=signature_params, documented_params=all_params
                )
            )

        return errs

    @property
    def directives_without_two_colons(self):
        return DIRECTIVE_PATTERN.findall(self.raw_doc)

    def parameter_type(self, param):
        return self.doc_all_parameters[param][0]

    @property
    def see_also(self):
        result = collections.OrderedDict()
        for funcs, desc in self.doc["See Also"]:
            for func, _ in funcs:
                result[func] = "".join(desc)

        return result

    @property
    def examples(self):
        return self.doc["Examples"]

    @property
    def returns(self):
        return self.doc["Returns"]

    @property
    def yields(self):
        return self.doc["Yields"]

    @property
    def method_source(self):
        try:
            source = inspect.getsource(self.obj)
        except TypeError:
            return ""
        return textwrap.dedent(source)

    @property
    def method_returns_something(self):
        """
        Check if the docstrings method can return something.

        Bare returns, returns valued None and returns from nested functions are
        disconsidered.

        Returns
        -------
        bool
            Whether the docstrings method can return something.
        """

        def get_returns_not_on_nested_functions(node):
            returns = [node] if isinstance(node, ast.Return) else []
            for child in ast.iter_child_nodes(node):
                # Ignore nested functions and its subtrees.
                if not isinstance(child, ast.FunctionDef):
                    child_returns = get_returns_not_on_nested_functions(child)
                    returns.extend(child_returns)
            return returns

        tree = ast.parse(self.method_source).body
        if tree:
            returns = get_returns_not_on_nested_functions(tree[0])
            return_values = [r.value for r in returns]
            # Replace Constant nodes valued None for None.
            for i, v in enumerate(return_values):
                if isinstance(v, ast.Constant) and v.value is None:
                    return_values[i] = None
            return any(return_values)
        else:
            return False

    @property
    def deprecated(self):
        return ".. deprecated:: " in (self.summary + self.extended_summary)


def _check_desc(desc, code_no_desc, code_no_upper, code_no_period, **kwargs):
    # Find and strip out any sphinx directives
    desc = "\n".join(desc)
    for directive in DIRECTIVES:
        full_directive = f".. {directive}"
        if full_directive in desc:
            # Only retain any description before the directive
            desc = desc[: desc.index(full_directive)].rstrip("\n")
    desc = desc.split("\n")

    errs = list()
    if not "".join(desc):
        errs.append(error(code_no_desc, **kwargs))
    else:
        if desc[0][0].isalpha() and not desc[0][0].isupper():
            errs.append(error(code_no_upper, **kwargs))
        # Not ending in "." is only an error if the last bit is not
        # indented (e.g., quote or code block)
        if not desc[-1].endswith(".") and not desc[-1].startswith(IGNORE_STARTS):
            errs.append(error(code_no_period, **kwargs))
    return errs


def _find_class_node(module_node: ast.AST, cls_name) -> ast.ClassDef:
    #  Find the class node within a module, when checking constructor docstrings.
    for node in ast.walk(module_node):
        if isinstance(node, ast.ClassDef) and node.name == cls_name:
            return node
    raise ValueError(f"Could not find class node {cls_name}")


def validate(obj_name, validator_cls=None, **validator_kwargs):
    """
    Validate the docstring.

    Parameters
    ----------
    obj_name : str
        The name of the object whose docstring will be evaluated, e.g.
        'pandas.read_csv'. The string must include the full, unabbreviated
        package/module names, i.e. 'pandas.read_csv', not 'pd.read_csv' or
        'read_csv'.
    validator_cls : Validator, optional
        The Validator class to use. By default, :class:`Validator` will be used.
    **validator_kwargs
        Keyword arguments to pass to ``validator_cls`` upon initialization.
        Note that ``obj_name`` will be passed as a named argument as well.

    Returns
    -------
    dict
        A dictionary containing all the information obtained from validating
        the docstring.

    Notes
    -----
    The errors codes are defined as:
    - First two characters: Section where the error happens:
       * GL: Global (no section, like section ordering errors)
       * SS: Short summary
       * ES: Extended summary
       * PR: Parameters
       * RT: Returns
       * YD: Yields
       * RS: Raises
       * WN: Warns
       * SA: See Also
       * NT: Notes
       * RF: References
       * EX: Examples
    - Last two characters: Numeric error code inside the section

    For example, PR02 is the second codified error in the Parameters section
    (which in this case is assigned to the error when unknown parameters are documented).

    The error codes, their corresponding error messages, and the details on how
    they are validated, are not documented more than in the source code of this
    function.
    """
    if not validator_cls:
        if isinstance(obj_name, str):
            doc = Validator(get_doc_object(Validator._load_obj(obj_name)))
        else:
            doc = Validator(obj_name)
    else:
        doc = validator_cls(obj_name=obj_name, **validator_kwargs)

    # lineno is only 0 if we have a module docstring in the file, and we are
    # validating that, so we change to 1 for readability of the output
    ignore_validation_comments = extract_ignore_validation_comments(
        doc.source_file_name
    ).get(doc.source_file_def_line or 1, [])

    errs = []
    if not doc.raw_doc:
        report_GL08: bool = True
        # Check if the object is a class and has a docstring in the constructor
        # Also check if code_obj is defined, as undefined for the AstValidator in validate_docstrings.py.
        if doc.is_function_or_method and doc.name.endswith(".__init__"):
            # Import here at runtime to avoid circular import as
            # AstValidator is a subclass of Validator class without `doc_obj` attribute.
            from numpydoc.hooks.validate_docstrings import (
                AstValidator,  # Support abstract syntax tree hook.
            )

            if hasattr(doc, "code_obj"):  # All Validator objects have this attr.
                cls_name = ".".join(
                    doc.code_obj.__qualname__.split(".")[:-1]
                )  # Collect all class depths before the constructor.
                cls = Validator._load_obj(f"{doc.code_obj.__module__}.{cls_name}")
                # cls = Validator._load_obj(f"{doc.name[:-9]}.{cls_name}") ## Alternative
                cls_doc = Validator(get_doc_object(cls))
            elif isinstance(doc, AstValidator):  # Supports class traversal for ASTs.
                ancestry = doc.ancestry
                if len(ancestry) > 2:  # e.g. module.class.__init__
                    parent = doc.ancestry[-1]  # Get the parent
                    cls_name = ".".join(
                        [
                            getattr(node, "name", node.__module__)
                            for node in doc.ancestry
                        ]
                    )
                    cls_doc = AstValidator(
                        ast_node=parent,
                        filename=doc.source_file_name,
                        obj_name=cls_name,
                        ancestry=doc.ancestry[:-1],
                    )
                else:
                    # Ignore edge case: __init__ functions that don't belong to a class.
                    cls_doc = None
            else:
                raise TypeError(
                    f"Cannot load {doc.name} as a usable Validator object (Validator does not have `doc_obj` attr or type `AstValidator`)."
                )

            # Parameter_mismatches, PR01, PR02, PR03 are checked for the class docstring.
            # If cls_doc has PR01, PR02, PR03 errors, i.e. invalid class docstring,
            # then we also report missing constructor docstring, GL08.
            if cls_doc:
                report_GL08 = len(cls_doc.parameter_mismatches) > 0

        # Check if GL08 is to be ignored:
        if "GL08" in ignore_validation_comments:
            report_GL08 = False
        # Add GL08 error?
        if report_GL08:
            errs.append(error("GL08"))
        return {
            "type": doc.type,
            "docstring": doc.clean_doc,
            "deprecated": doc.deprecated,
            "file": doc.source_file_name,
            "file_line": doc.source_file_def_line,
            "errors": errs,
            "examples_errors": "",
        }

    if doc.start_blank_lines != 1 and "\n" in doc.raw_doc:
        errs.append(error("GL01"))
    if doc.end_blank_lines != 1 and "\n" in doc.raw_doc:
        errs.append(error("GL02"))
    if doc.double_blank_lines:
        errs.append(error("GL03"))
    for line in doc.raw_doc.splitlines():
        if re.match("^ *\t", line):
            errs.append(error("GL05", line_with_tabs=line.lstrip()))

    unexpected_sections = [
        section for section in doc.section_titles if section not in ALLOWED_SECTIONS
    ]
    for section in unexpected_sections:
        errs.append(
            error("GL06", section=section, allowed_sections=", ".join(ALLOWED_SECTIONS))
        )

    correct_order = [
        section for section in ALLOWED_SECTIONS if section in doc.section_titles
    ]
    if correct_order != doc.section_titles:
        errs.append(error("GL07", correct_sections=", ".join(correct_order)))

    if doc.deprecated and not doc.extended_summary.startswith(".. deprecated:: "):
        errs.append(error("GL09"))

    directives_without_two_colons = doc.directives_without_two_colons
    if directives_without_two_colons:
        errs.append(error("GL10", directives=directives_without_two_colons))

    if not doc.summary:
        errs.append(error("SS01"))
    else:
        if doc.summary[0].isalpha() and not doc.summary[0].isupper():
            errs.append(error("SS02"))
        if doc.summary[-1] != ".":
            errs.append(error("SS03"))
        if doc.summary != doc.summary.lstrip():
            errs.append(error("SS04"))
        # Heuristic to check for infinitive verbs - shouldn't end in "s"
        elif (
            doc.is_function_or_method
            and len(doc.summary.split(" ")[0]) > 1
            and doc.summary.split(" ")[0][-1] == "s"
            and doc.summary.split(" ")[0][-2] != "s"
        ):
            errs.append(error("SS05"))
        if doc.num_summary_lines > 1:
            errs.append(error("SS06"))

    if not doc.is_mod and not doc.extended_summary:
        errs.append(("ES01", "No extended summary found"))

    # PR01: Parameters not documented
    # PR02: Unknown parameters
    # PR03: Wrong parameters order
    errs += doc.parameter_mismatches

    for param, kind_desc in doc.doc_all_parameters.items():
        if not param.startswith("*"):  # Check can ignore var / kwargs
            if not doc.parameter_type(param):
                if ":" in param:
                    errs.append(error("PR10", param_name=param.split(":")[0]))
                else:
                    errs.append(error("PR04", param_name=param))
            else:
                if doc.parameter_type(param)[-1] == ".":
                    errs.append(error("PR05", param_name=param))
                # skip common_type_error checks when the param type is a set of
                # options
                if "{" in doc.parameter_type(param):
                    continue
                common_type_errors = [
                    ("integer", "int"),
                    ("boolean", "bool"),
                    ("string", "str"),
                ]
                for wrong_type, right_type in common_type_errors:
                    if wrong_type in set(re.split(r"\W", doc.parameter_type(param))):
                        errs.append(
                            error(
                                "PR06",
                                param_name=param,
                                right_type=right_type,
                                wrong_type=wrong_type,
                            )
                        )
        errs.extend(_check_desc(kind_desc[1], "PR07", "PR08", "PR09", param_name=param))

    if doc.is_function_or_method:
        if not doc.returns:
            if doc.method_returns_something:
                errs.append(error("RT01"))
        else:
            if len(doc.returns) == 1 and doc.returns[0].name:
                errs.append(error("RT02"))
            for name_or_type, type_, desc in doc.returns:
                errs.extend(_check_desc(desc, "RT03", "RT04", "RT05"))

        if not doc.yields and doc.is_generator_function:
            errs.append(error("YD01"))

    if not doc.is_mod:
        if not doc.see_also:
            errs.append(error("SA01"))
        else:
            for rel_name, rel_desc in doc.see_also.items():
                if rel_desc:
                    if not rel_desc.endswith("."):
                        errs.append(error("SA02", reference_name=rel_name))
                    if rel_desc[0].isalpha() and not rel_desc[0].isupper():
                        errs.append(error("SA03", reference_name=rel_name))
                else:
                    errs.append(error("SA04", reference_name=rel_name))

        if not doc.examples:
            errs.append(error("EX01"))

    errs = [err for err in errs if err[0] not in ignore_validation_comments]

    return {
        "type": doc.type,
        "docstring": doc.clean_doc,
        "deprecated": doc.deprecated,
        "file": doc.source_file_name,
        "file_line": doc.source_file_def_line,
        "errors": errs,
    }



import inspect
import io
import sys

import pytest

import numpydoc
import numpydoc.cli


def _capture_stdout(func_name, *args, **kwargs):
    """
    Return stdout of calling `func_name`.

    This docstring should be perfect, as it is used to test the
    validation with a docstring without errors.

    Parameters
    ----------
    func_name : callable
        Function to be called.
    *args, **kwargs
        Will be passed to `func_name`.

    Returns
    -------
    str
        The content that the function printed.

    See Also
    --------
    sys.stdout : Python's file handler for stdout.

    Examples
    --------
    >>> _capture_stdout(print, "hello world")
    'hello world'
    """
    f = io.StringIO()
    sys.stdout, old_stdout = f, sys.stdout
    try:
        func_name(*args, **kwargs)
        return f.getvalue().strip("\n\r")
    finally:
        sys.stdout = old_stdout


def _docstring_with_errors():
    """
    this docstring should report some errors

    Parameters
    ----------
    made_up_param : str
    """


def _invalid_docstring():
    """
    This docstring should break the parsing.

    See Also
    --------
    : this is invalid
    """


def test_renders_package_docstring():
    out = _capture_stdout(numpydoc.cli.render_object, "numpydoc")
    assert out.startswith("This package provides the numpydoc Sphinx")


def test_renders_module_docstring(capsys):
    numpydoc.cli.main(["render", "numpydoc.cli"])
    out = capsys.readouterr().out.strip("\n\r")
    assert out.startswith(numpydoc.cli.__doc__)


def test_renders_function_docstring():
    out = _capture_stdout(
        numpydoc.cli.render_object, "numpydoc.tests.test_main._capture_stdout"
    )
    assert out.startswith("Return stdout of calling")


def test_render_object_returns_correct_exit_status():
    exit_status = numpydoc.cli.render_object("numpydoc.tests.test_main._capture_stdout")
    assert exit_status == 0

    with pytest.raises(ValueError):
        numpydoc.cli.render_object("numpydoc.tests.test_main._invalid_docstring")


def test_validate_detects_errors():
    out = _capture_stdout(
        numpydoc.cli.validate_object,
        "numpydoc.tests.test_main._docstring_with_errors",
    )
    assert "SS02" in out
    assert "Summary does not start with a capital letter" in out

    exit_status = numpydoc.cli.validate_object(
        "numpydoc.tests.test_main._docstring_with_errors"
    )
    assert exit_status > 0


def test_validate_perfect_docstring():
    out = _capture_stdout(
        numpydoc.cli.validate_object, "numpydoc.tests.test_main._capture_stdout"
    )
    assert out == ""

    exit_status = numpydoc.cli.validate_object(
        "numpydoc.tests.test_main._capture_stdout"
    )
    assert exit_status == 0


@pytest.mark.parametrize("args", [[], ["--ignore", "SS03"]])
def test_lint(capsys, args):
    argv = ["lint", "numpydoc/__main__.py"] + args
    if args:
        expected = ""
        expected_status = 0
    else:
        expected = "numpydoc/__main__.py:1: SS03 Summary does not end with a period"
        expected_status = 1

    return_status = numpydoc.cli.main(argv)
    err = capsys.readouterr().err.strip("\n\r")
    assert err == expected
    assert return_status == expected_status


def test_lint_help(capsys):
    """Test that lint help section is displaying."""

    with pytest.raises(SystemExit):
        return_code = numpydoc.cli.main(["lint", "--help"])
        assert return_code == 0

    out = capsys.readouterr().out
    assert "--ignore" in out
    assert "--config" in out



"""Test utility functions for hooks."""

from pathlib import Path

import pytest

from numpydoc.hooks import utils


@pytest.mark.parametrize(
    ["reason_file", "files", "expected_reason"],
    [
        (None, None, "current directory"),
        (None, ["x.py"], "file system root"),
        (".git", ["x.py"], "version control"),
        ("pyproject.toml", ["x.py"], "pyproject.toml"),
        ("setup.cfg", ["x.py"], "setup.cfg"),
    ],
)
def test_find_project_root(tmp_path, request, reason_file, files, expected_reason):
    """Test the process of finding the project root."""
    if reason_file:
        (tmp_path / reason_file).touch()

    if files:
        expected_dir = (
            Path(tmp_path.anchor) if expected_reason == "file system root" else tmp_path
        )

        for file in files:
            (tmp_path / file).touch()
    else:
        expected_dir = request.config.rootdir

    root_dir, reason = utils.find_project_root(files if not files else [tmp_path])
    assert reason == expected_reason
    assert root_dir == expected_dir



"""Run numpydoc validation on contents of a file."""

import ast
import configparser
import os
import re
import sys

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

from .. import docscrape, validate
from .utils import find_project_root


class AstValidator(validate.Validator):
    """
    Overrides the :class:`Validator` to work entirely with the AST.

    Parameters
    ----------
    ast_node : ast.AST
        The node under inspection.
    filename : os.PathLike
        The file where the node is defined.
    obj_name : str
        A name for the node to use in the listing of issues for the file as a whole.
    """

    def __init__(
        self,
        *,
        ast_node: ast.AST,
        filename: os.PathLike,
        obj_name: str,
        ancestry: list[ast.AST],
    ) -> None:
        self.node: ast.AST = ast_node
        self.raw_doc: str = ast.get_docstring(self.node, clean=False) or ""
        self.clean_doc: str = ast.get_docstring(self.node, clean=True)
        self.doc: docscrape.NumpyDocString = docscrape.NumpyDocString(self.raw_doc)

        self._source_file: os.PathLike = Path(filename).resolve()
        self._name: str = obj_name

        self.is_class: bool = isinstance(ast_node, ast.ClassDef)
        self.is_module: bool = isinstance(ast_node, ast.Module)

        self.ancestry: list[ast.AST] = ancestry

    @staticmethod
    def _load_obj(name):
        raise NotImplementedError("AstValidator does not support this method.")

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_function_or_method(self) -> bool:
        return isinstance(self.node, ast.FunctionDef | ast.AsyncFunctionDef)

    @property
    def is_mod(self) -> bool:
        return self.is_module

    @property
    def is_generator_function(self) -> bool:
        if not self.is_function_or_method:
            return False
        for child in ast.iter_child_nodes(self.node):
            if isinstance(child, ast.Expr) and isinstance(child.value, ast.Yield):
                return True
        return False

    @property
    def type(self) -> str:
        if self.is_function_or_method:
            return "function"
        if self.is_class:
            return "type"
        if self.is_module:
            return "module"
        raise ValueError("Unknown type.")

    @property
    def source_file_name(self) -> str:
        return self._source_file

    @property
    def source_file_def_line(self) -> int:
        return self.node.lineno if not self.is_module else 1

    @property
    def signature_parameters(self) -> Tuple[str]:
        def extract_signature(node, parent):
            args_node = node.args
            params = []
            for arg_type in ["posonlyargs", "args", "vararg", "kwonlyargs", "kwarg"]:
                entries = getattr(args_node, arg_type)
                if arg_type in ["vararg", "kwarg"]:
                    if entries and arg_type == "vararg":
                        params.append(f"*{entries.arg}")
                    if entries and arg_type == "kwarg":
                        params.append(f"**{entries.arg}")
                else:
                    params.extend([arg.arg for arg in entries])
            params = tuple(params)
            if (
                params
                and params[0] in {"self", "cls"}
                and isinstance(parent, ast.ClassDef)
            ):
                return params[1:]
            return params

        params = tuple()
        if self.is_function_or_method:
            params = extract_signature(self.node, self.ancestry[-1])
        elif self.is_class:
            for child in self.node.body:
                if isinstance(child, ast.FunctionDef) and child.name == "__init__":
                    params = extract_signature(child, self.node)
        return params

    @property
    def method_source(self) -> str:
        with open(self.source_file_name) as file:
            source = ast.get_source_segment(file.read(), self.node)
        return source


class DocstringVisitor(ast.NodeVisitor):
    """
    Visits nodes in the AST from a given module and reporting numpydoc issues.

    Parameters
    ----------
    filepath : str
        The absolute or relative path to the file to inspect.
    config : dict
        Configuration options for reviewing flagged issues.
    """

    def __init__(
        self,
        filepath: str,
        config: dict,
    ) -> None:
        self.config: dict = config
        self.filepath: str = filepath
        self.module_name: str = Path(self.filepath).stem
        self.stack: list[ast.AST] = []
        self.findings: list = []

    @property
    def node_name(self) -> str:
        """
        Get the full name of the current node in the stack.

        Returns
        -------
        str
            The full name of the current node in the stack.
        """
        return ".".join(
            [getattr(node, "name", self.module_name) for node in self.stack]
        )

    def _ignore_issue(self, node: ast.AST, check: str) -> bool:
        """
        Check whether the issue should be ignored.

        Parameters
        ----------
        node : ast.AST
            The node under inspection.
        check : str
            The code for the check being evaluated.

        Return
        ------
        bool
            Whether the issue should be excluded from the report.
        """
        if check not in self.config["checks"]:
            return True

        if self.config["overrides"]:
            try:
                pattern = self.config["overrides"][check]
                if re.search(pattern, ast.get_docstring(node)) is not None:
                    return True
            except KeyError:
                pass

        return False

    def _get_numpydoc_issues(self, node: ast.AST) -> None:
        """
        Get numpydoc validation issues.

        Parameters
        ----------
        node : ast.AST
            The node under inspection.
        """
        name = self.node_name
        report = validate.validate(
            name,
            AstValidator,
            ast_node=node,
            filename=self.filepath,
            ancestry=self.stack[:-1],
        )
        self.findings.extend(
            [
                [f"{self.filepath}:{report['file_line']}", name, check, description]
                for check, description in report["errors"]
                if not self._ignore_issue(node, check)
            ]
        )

    def visit(self, node: ast.AST) -> None:
        """
        Visit a node in the AST and report on numpydoc validation issues.

        Parameters
        ----------
        node : ast.AST
            The node to visit.
        """
        if isinstance(
            node, ast.Module | ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef
        ):
            self.stack.append(node)

            if not (
                self.config["exclude"]
                and re.search(self.config["exclude"], self.node_name)
            ):
                self._get_numpydoc_issues(node)

            self.generic_visit(node)
            _ = self.stack.pop()


def parse_config(dir_path: os.PathLike = None) -> dict:
    """
    Parse config information from a pyproject.toml or setup.cfg file.

    This function looks in the provided directory path first for a
    pyproject.toml file. If it finds that, it won't look for a setup.cfg
    file.

    Parameters
    ----------
    dir_path : os.PathLike
        An absolute or relative path to a directory containing
        either a pyproject.toml file specifying a
        [tool.numpydoc_validation] section or a setup.cfg file
        specifying a [tool:numpydoc_validation] section.
        For example, ``~/my_project``. If not provided, the hook
        will try to find the project root directory.

    Returns
    -------
    dict
        Config options for the numpydoc validation hook.
    """
    options = {
        "checks": {"all"},
        "exclude": set(),
        "overrides": {},
        "exclude_files": set(),
    }
    dir_path = Path(dir_path).expanduser().resolve()

    toml_path = dir_path / "pyproject.toml"
    cfg_path = dir_path / "setup.cfg"

    def compile_regex(expressions):
        return (
            re.compile(r"|".join(exp for exp in expressions if exp))
            if expressions
            else None
        )

    def extract_check_overrides(options, config_items):
        for option, value in config_items:
            if option.startswith("override_"):
                _, check = option.split("_")
                if value:
                    options["overrides"][check.upper()] = compile_regex(value)

    if toml_path.is_file():
        with open(toml_path, "rb") as toml_file:
            pyproject_toml = tomllib.load(toml_file)
            config = pyproject_toml.get("tool", {}).get("numpydoc_validation", {})
            options["checks"] = set(config.get("checks", options["checks"]))

            global_exclusions = config.get("exclude", options["exclude"])
            options["exclude"] = set(
                global_exclusions
                if not isinstance(global_exclusions, str)
                else [global_exclusions]
            )

            file_exclusions = config.get("exclude_files", options["exclude_files"])
            options["exclude_files"] = set(
                file_exclusions
                if not isinstance(file_exclusions, str)
                else [file_exclusions]
            )

            extract_check_overrides(options, config.items())

    elif cfg_path.is_file():
        config = configparser.ConfigParser()
        config.read(cfg_path)
        numpydoc_validation_config_section = "tool:numpydoc_validation"
        try:
            try:
                options["checks"] = set(
                    config.get(numpydoc_validation_config_section, "checks")
                    .rstrip(",")
                    .split(",")
                    or options["checks"]
                )
            except configparser.NoOptionError:
                pass
            try:
                options["exclude"] = set(
                    config.get(numpydoc_validation_config_section, "exclude")
                    .rstrip(",")
                    .split(",")
                    or options["exclude"]
                )
            except configparser.NoOptionError:
                pass

            try:
                options["exclude_files"] = set(
                    config.get(numpydoc_validation_config_section, "exclude_files")
                    .rstrip(",")
                    .split(",")
                    or options["exclude_files"]
                )
            except configparser.NoOptionError:
                pass

            extract_check_overrides(
                options, config.items(numpydoc_validation_config_section)
            )

        except configparser.NoSectionError:
            pass

    options["checks"] = validate.get_validation_checks(options["checks"])
    options["exclude"] = compile_regex(options["exclude"])
    options["exclude_files"] = compile_regex(options["exclude_files"])
    return options


def process_file(filepath: os.PathLike, config: dict) -> "list[list[str]]":
    """
    Run numpydoc validation on a file.

    Parameters
    ----------
    filepath : path-like
        The absolute or relative path to the file to inspect.
    config : dict
        Configuration options for reviewing flagged issues.

    Returns
    -------
    list[list[str]]
        A list of [name, check, description] lists for flagged issues.
    """
    with open(filepath) as file:
        module_node = ast.parse(file.read(), filepath)

    docstring_visitor = DocstringVisitor(filepath=str(filepath), config=config)
    docstring_visitor.visit(module_node)

    return docstring_visitor.findings


def run_hook(
    files: List[str],
    *,
    config: Dict[str, Any] | None = None,
    ignore: List[str] | None = None,
) -> int:
    """
    Run the numpydoc validation hook.

    Parameters
    ----------
    files : list[str]
        The absolute or relative paths to the files to inspect.
    config : Union[dict[str, Any], None], optional
        Configuration options for reviewing flagged issues.
    ignore : Union[list[str], None], optional
        Checks to ignore in the results.

    Returns
    -------
    int
        The return status: 1 if issues were found, 0 otherwise.
    """
    project_root, _ = find_project_root(files)
    config_options = parse_config(config or project_root)
    config_options["checks"] -= set(ignore or [])
    exclude_re = config_options["exclude_files"]

    findings = False
    for file in files:
        if exclude_re and exclude_re.match(file):
            continue
        if file_issues := process_file(file, config_options):
            findings = True

            for line, obj, check, description in file_issues:
                print(f"\n{line}: {check} {description}", file=sys.stderr)

    return int(findings)



import warnings
from contextlib import nullcontext
from dataclasses import dataclass
from functools import cached_property, partial, wraps
from inspect import getsourcefile, getsourcelines

import pytest

import numpydoc.tests
from numpydoc import validate
from numpydoc.docscrape import get_doc_object
from numpydoc.validate import Validator

validate_one = validate.validate

ALL_CHECKS = set(validate.ERROR_MSGS.keys())


@pytest.mark.parametrize(
    ["checks", "expected"],
    [
        [{"all"}, ALL_CHECKS],
        [set(), set()],
        [{"EX01"}, {"EX01"}],
        [{"EX01", "SA01"}, {"EX01", "SA01"}],
        [{"all", "EX01", "SA01"}, ALL_CHECKS - {"EX01", "SA01"}],
        [{"all", "PR01"}, ALL_CHECKS - {"PR01"}],
    ],
)
def test_utils_get_validation_checks(checks, expected):
    """Ensure check selection is working."""
    assert validate.get_validation_checks(checks) == expected


@pytest.mark.parametrize(
    "checks",
    [
        {"every"},
        {None},
        {"SM10"},
        {"EX01", "SM10"},
    ],
)
def test_get_validation_checks_validity(checks):
    """Ensure that invalid checks are flagged."""
    with pytest.raises(ValueError, match="Unrecognized validation code"):
        _ = validate.get_validation_checks(checks)


class _DummyList(list):
    """Dummy list class to test validation."""


def test_no_file():
    """Test that validation can be done on functions made on the fly."""
    # Just a smoke test for now, <list> will have a None filename
    validate.validate("numpydoc.tests.test_validate._DummyList.clear")


@pytest.mark.parametrize(
    ["file_contents", "expected"],
    [
        ["class MyClass:\n    pass", {}],
        ["class MyClass:  # numpydoc ignore=EX01\n    pass", {1: ["EX01"]}],
        [
            "class MyClass:  # numpydoc ignore= EX01,SA01\n    pass",
            {1: ["EX01", "SA01"]},
        ],
        [
            "class MyClass:\n    def my_method():  # numpydoc ignore:EX01\n        pass",
            {2: ["EX01"]},
        ],
        [
            "class MyClass:\n    def my_method():  # numpydoc ignore: EX01,PR01\n        pass",
            {2: ["EX01", "PR01"]},
        ],
        [
            "class MyClass:  # numpydoc ignore=GL08\n    def my_method():  # numpydoc ignore:EX01,PR01\n        pass",
            {1: ["GL08"], 2: ["EX01", "PR01"]},
        ],
    ],
)
def test_extract_ignore_validation_comments(tmp_path, file_contents, expected):
    """Test that extraction of validation ignore comments is working."""
    filepath = tmp_path / "ignore_comments.py"
    with open(filepath, "w") as file:
        file.write(file_contents)
    assert validate.extract_ignore_validation_comments(filepath) == expected


@pytest.mark.parametrize(
    "assumed_encoding",
    (
        pytest.param("utf-8", id="utf8_codec"),
        pytest.param("cp1252", id="cp1252_codec"),
    ),
)
@pytest.mark.parametrize(
    ("classname", "actual_encoding"),
    (
        pytest.param("MÿClass", "cp1252", id="cp1252_file"),
        pytest.param("My\u0081Class", "utf-8", id="utf8_file"),
    ),
)
def test_encodings(tmp_path, classname, actual_encoding, assumed_encoding):
    """Test handling of different source file encodings."""
    # write file as bytes with `actual_encoding`
    filepath = tmp_path / "ignore_comments.py"
    file_contents = f"class {classname}:\n    pass"
    with open(filepath, "wb") as file:
        file.write(file_contents.encode(actual_encoding))
    # this should fail on the ÿ in MÿClass. It represents the (presumed rare) case where
    # a user's editor saved the source file in cp1252 (or anything other than utf-8).
    if actual_encoding == "cp1252" and assumed_encoding == "utf-8":
        context = partial(
            pytest.raises,
            UnicodeDecodeError,
            match="can't decode byte 0xff in position 7: invalid start byte",
        )
    # this is the more likely case: file was utf-8 encoded, but Python on Windows uses
    # the system codepage to read the file. This case is fixed by numpy/numpydoc#510
    elif actual_encoding == "utf-8" and assumed_encoding == "cp1252":
        context = partial(
            pytest.raises,
            UnicodeDecodeError,
            match="can't decode byte 0x81 in position 9: character maps to <undefined>",
        )
    else:
        context = nullcontext
    with context():
        result = validate.extract_ignore_validation_comments(filepath, assumed_encoding)
        assert result == {}


class GoodDocStrings:
    """
    Collection of good doc strings.

    This class contains a lot of docstrings that should pass the validation
    script without any errors.

    See Also
    --------
    AnotherClass : With its description.

    Examples
    --------
    >>> result = 1 + 1
    """

    def one_liner(self):
        """Allow one liner docstrings (including quotes)."""
        # This should fail, but not because of the position of the quotes

    def plot(self, kind, color="blue", **kwargs):
        """
        Generate a plot.

        Render the data in the Series as a matplotlib plot of the
        specified kind.

        Parameters
        ----------
        kind : str
            Kind of matplotlib plot, e.g.::

                "foo"

        color : str, default 'blue'
            Color name or rgb code.
        **kwargs
            These parameters will be passed to the matplotlib plotting
            function.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def swap(self, arr, i, j, *args, **kwargs):
        """
        Swap two indices on an array.

        The extended summary can be multiple paragraphs, but just one
        is enough to pass the validation.

        Parameters
        ----------
        arr : list
            The list having indexes swapped.
        i, j : int
            The indexes being swapped.
        *args, **kwargs
            Extraneous parameters are being permitted.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def sample(self):
        """
        Generate and return a random number.

        The value is sampled from a continuous uniform distribution between
        0 and 1.

        Returns
        -------
        float
            Random number generated.

            - Make sure you set a seed for reproducibility

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def random_letters(self):
        """
        Generate and return a sequence of random letters.

        The length of the returned string is also random, and is also
        returned.

        Returns
        -------
        length : int
            Length of the returned string.
        letters : str
            String of random letters.

            .. versionadded:: 0.1

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def sample_values(self):
        """
        Generate an infinite sequence of random numbers.

        The values are sampled from a continuous uniform distribution between
        0 and 1.

        Yields
        ------
        float
            Random number generated.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def head(self):
        """
        Return the first 5 elements of the Series.

        This function is mainly useful to preview the values of the
        Series without displaying the whole of it.

        Returns
        -------
        int
            Subset of the original series with the 5 first values.

        See Also
        --------
        Series.tail : Return the last 5 elements of the Series.
        Series.iloc : Return a slice of the elements in the Series,
            which can also be used to return the first or last n.

        Examples
        --------
        >>> 1 + 1
        2
        """
        return 1

    def head1(self, n=5):
        """
        Return the first elements of the Series.

        This function is mainly useful to preview the values of the
        Series without displaying the whole of it.

        Parameters
        ----------
        n : int
            Number of values to return.

        Returns
        -------
        int
            Subset of the original series with the n first values.

        See Also
        --------
        tail : Return the last n elements of the Series.

        Examples
        --------
        >>> s = 10
        >>> s
        10

        With the `n` parameter, we can change the number of returned rows:

        >>> s + 1
        11
        """
        return 1

    def summary_starts_with_number(self, n=5):
        """
        2nd rule of summaries should allow this.

        3 Starting the summary with a number instead of a capital letter.
        Also in parameters, returns, see also...

        Parameters
        ----------
        n : int
            4 Number of values to return.

        Returns
        -------
        int
            5 Subset of the original series with the n first values.

        See Also
        --------
        tail : 6 Return the last n elements of the Series.

        Examples
        --------
        >>> s = 10
        >>> s
        10

        7 With the `n` parameter, we can change the number of returned rows:

        >>> s + 1
        11
        """
        return 1

    def contains(self, pat, case=True, na=float("NaN")):
        """
        Return whether each value contains `pat`.

        In this case, we are illustrating how to use sections, even
        if the example is simple enough and does not require them.

        Parameters
        ----------
        pat : str
            Pattern to check for within each element.
        case : bool, default True
            Whether check should be done with case sensitivity.
        na : object, default np.nan
            Fill value for missing data.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> s = 25
        >>> s
        25

        **Case sensitivity**

        With `case_sensitive` set to `False` we can match `a` with both
        `a` and `A`:

        >>> s + 1
        26

        **Missing values**

        We can fill missing values in the output using the `na` parameter:

        >>> s * 2
        50
        """

    def mode(self, axis, numeric_only):
        """
        Ensure reST directives don't affect checks for leading periods.

        The extended summary can be multiple paragraphs, but just one
        is enough to pass the validation.

        Parameters
        ----------
        axis : str
            Sentence ending in period, followed by single directive.

            .. versionchanged:: 0.1.2

        numeric_only : bool
            Sentence ending in period, followed by multiple directives.

            .. versionadded:: 0.1.2
            .. deprecated:: 0.00.0
                A multiline description,
                which spans another line.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def good_imports(self):
        """
        Ensure import other than numpy and pandas are fine.

        The extended summary can be multiple paragraphs, but just one
        is enough to pass the validation.

        See Also
        --------
        related : Something related.

        Examples
        --------
        This example does not import pandas or import numpy.
        >>> import datetime
        >>> datetime.MAXYEAR
        9999
        """

    def no_returns(self):
        """
        Say hello and have no returns.

        The extended summary can be multiple paragraphs, but just one
        is enough to pass the validation.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def empty_returns(self):
        """
        Say hello and always return None.

        Since this function never returns a value, this
        docstring doesn't need a return section.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

        def say_hello():
            return "Hello World!"

        say_hello()
        if True:
            return
        else:
            return

    def warnings(self):
        """
        Do one thing.

        Sometimes, this function does other things.

        Warnings
        --------
        This function may produce side effects when some condition
        is met.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def multiple_variables_on_one_line(self, matrix, a, b, i, j):
        """
        Swap two values in a matrix.

        The extended summary can be multiple paragraphs, but just one
        is enough to pass the validation.

        Parameters
        ----------
        matrix : list of list
            A double list that represents a matrix.
        a, b : int
            The indices of the first value.
        i, j : int
            The indices of the second value.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def other_parameters(self, param1, param2):
        """
        Ensure "Other Parameters" are recognized.

        The second parameter is used infrequently, so it is put in the
        "Other Parameters" section.

        Parameters
        ----------
        param1 : bool
            Description of commonly used parameter.

        Other Parameters
        ----------------
        param2 : str
            Description of infrequently used parameter.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def valid_options_in_parameter_description_sets(self, bar):
        """
        Ensure a PR06 error is not raised when type is member of a set.

        Literal keywords like 'integer' are valid when specified in a set of
        valid options for a keyword parameter.

        Parameters
        ----------
        bar : {'integer', 'boolean'}
            The literal values of 'integer' and 'boolean' are part of an
            options set and thus should not be subject to PR06 warnings.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def parameters_with_trailing_underscores(self, str_):
        r"""
        Ensure PR01 and PR02 errors are not raised with trailing underscores.

        Parameters with trailing underscores need to be escaped to render
        properly in the documentation since trailing underscores are used to
        create links. Doing so without also handling the change in the validation
        logic makes it impossible to both pass validation and render correctly.

        Parameters
        ----------
        str\_ : str
           Some text.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """

    def parameter_with_wrong_types_as_substrings(self, a, b, c, d, e, f):
        r"""
        Ensure PR06 doesn't fail when non-preferable types are substrings.

        While PR06 checks for parameter types which contain non-preferable type
        names like integer (int), string (str), and boolean (bool), PR06 should
        not fail if those types are used only as susbtrings in, for example,
        custom type names.

        Parameters
        ----------
        a : Myint
           Some text.
        b : intClass
           Some text.
        c : Mystring
           Some text.
        d : stringClass
           Some text.
        e : Mybool
           Some text.
        f : boolClass
           Some text.

        See Also
        --------
        related : Something related.

        Examples
        --------
        >>> result = 1 + 1
        """


class BadGenericDocStrings:
    """Everything here has a bad docstring"""

    def func(self):
        """Some function.

        With several mistakes in the docstring.

        It has a blank like after the signature `def func():`.

        The text 'Some function' should go in the line after the
        opening quotes of the docstring, not in the same line.

        There is a blank line between the docstring and the first line
        of code `foo = 1`.

        The closing quotes should be in the next line, not in this one."""

        foo = 1
        bar = 2
        return foo + bar

    def astype(self, dtype):
        """
        Casts Series type.

        Verb in third-person of the present simple, should be infinitive.
        """

    def astype1(self, dtype):
        """
        Method to cast Series type.

        Does not start with verb.
        """

    def astype2(self, dtype):
        """
        Cast Series type

        Missing dot at the end.
        """

    def astype3(self, dtype):
        """
        Cast Series type from its current type to the new type defined in
        the parameter dtype.

        Summary is too verbose and doesn't fit in a single line.
        """

    def two_linebreaks_between_sections(self, foo):
        """
        Test linebreaks message GL03.

        Note 2 blank lines before parameters section.


        Parameters
        ----------
        foo : str
            Description of foo parameter.
        """

    def linebreak_at_end_of_docstring(self, foo):
        """
        Test linebreaks message GL03.

        Note extra blank line at end of docstring.

        Parameters
        ----------
        foo : str
            Description of foo parameter.

        """

    def plot(self, kind, **kwargs):
        """
        Generate a plot.

        Render the data in the Series as a matplotlib plot of the
        specified kind.

        Note the blank line between the parameters title and the first
        parameter. Also, note that after the name of the parameter `kind`
        and before the colon, a space is missing.

        Also, note that the parameter descriptions do not start with a
        capital letter, and do not finish with a dot.

        Finally, the `**kwargs` parameter is missing.

        Parameters
        ----------

        kind: str
            kind of matplotlib plot
        """

    def unknown_section(self):
        """
        This section has an unknown section title.

        Unknown Section
        ---------------
        This should raise an error in the validation.
        """

    def sections_in_wrong_order(self):
        """
        This docstring has the sections in the wrong order.

        Parameters
        ----------
        name : str
            This section is in the right position.

        Examples
        --------
        >>> print("So far Examples is good, as it goes before Parameters")
        So far Examples is good, as it goes before Parameters

        See Also
        --------
        function : This should generate an error, as See Also needs to go
            before Examples.
        """

    def deprecation_in_wrong_order(self):
        """
        This docstring has the deprecation warning in the wrong order.

        This is the extended summary. The correct order should be
        summary, deprecation warning, extended summary.

        .. deprecated:: 1.0
            This should generate an error as it needs to go before
            extended summary.
        """

    def method_wo_docstrings(self):
        pass

    def directives_without_two_colons(self, first, second):
        """
        Ensure reST directives have trailing colons.

        Parameters
        ----------
        first : str
            Sentence ending in period, followed by single directive w/o colons.

            .. versionchanged 0.1.2

        second : bool
            Sentence ending in period, followed by multiple directives w/o
            colons.

            .. versionadded 0.1.2
            .. deprecated 0.00.0

        """


class WarnGenericFormat:
    """
    Those contains things that _may_ be incorrect formatting.
    """

    def too_short_header_underline(self, a, b):
        """
        The header line is too short.

        Parameters
        ------
        a, b : int
            Foo bar baz.
        """


class BadSummaries:
    def no_summary(self):
        """
        Returns
        -------
        int
            Always one.
        """

    def heading_whitespaces(self):
        """
           Summary with heading whitespaces.

        Returns
        -------
        int
            Always one.
        """

    def wrong_line(self):
        """Quotes are on the wrong line.

        Both opening and closing."""

    def no_punctuation(self):
        """
        Has the right line but forgets punctuation
        """

    def no_capitalization(self):
        """
        provides a lowercase summary.
        """

    def no_infinitive(self):
        """
        Started with a verb that is not infinitive.
        """

    def multi_line(self):
        """
        Extends beyond one line
        which is not correct.
        """

    def two_paragraph_multi_line(self):
        """
        Extends beyond one line
        which is not correct.

        Extends beyond one line, which in itself is correct but the
        previous short summary should still be an issue.
        """


class BadParameters:
    """
    Everything here has a problem with its Parameters section.
    """

    def no_type(self, value):
        """
        Lacks the type.

        Parameters
        ----------
        value
            A parameter without type.
        """

    def type_with_period(self, value):
        """
        Has period after type.

        Parameters
        ----------
        value : str.
            A parameter type should not finish with period.
        """

    def no_description(self, value):
        """
        Lacks the description.

        Parameters
        ----------
        value : str
        """

    def missing_params(self, kind, **kwargs):
        """
        Lacks kwargs in Parameters.

        Parameters
        ----------
        kind : str
            Foo bar baz.
        """

    def bad_colon_spacing(self, kind):
        """
        Has bad spacing in the type line.

        Parameters
        ----------
        kind: str
            Needs a space after kind.
        """

    def no_description_period(self, kind):
        """
        Forgets to add a period to the description.

        Parameters
        ----------
        kind : str
           Doesn't end with a dot
        """

    def no_description_period_with_directive(self, kind):
        """
        Forgets to add a period, and also includes a directive.

        Parameters
        ----------
        kind : str
           Doesn't end with a dot

           .. versionadded:: 0.00.0
        """

    def no_description_period_with_directives(self, kind):
        """
        Forgets to add a period, and also includes multiple directives.

        Parameters
        ----------
        kind : str
           Doesn't end with a dot

           .. versionchanged:: 0.00.0
           .. deprecated:: 0.00.0
        """

    def parameter_capitalization(self, kind):
        """
        Forgets to capitalize the description.

        Parameters
        ----------
        kind : str
           this is not capitalized.
        """

    def blank_lines(self, kind):
        """
        Adds a blank line after the section header.

        Parameters
        ----------

        kind : str
            Foo bar baz.
        """

    def integer_parameter(self, kind):
        """
        Uses integer instead of int.

        Parameters
        ----------
        kind : integer
            Foo bar baz.
        """

    def string_parameter(self, kind):
        """
        Uses string instead of str.

        Parameters
        ----------
        kind : string
            Foo bar baz.
        """

    def boolean_parameter(self, kind):
        """
        Uses boolean instead of bool.

        Parameters
        ----------
        kind : boolean
            Foo bar baz.
        """

    def list_incorrect_parameter_type(self, kind):
        """
        Uses list of boolean instead of list of bool.

        Parameters
        ----------
        kind : list of boolean, integer, float or string
            Foo bar baz.
        """

    def bad_parameter_spacing(self, a, b):
        """
        The parameters on the same line have an extra space between them.

        Parameters
        ----------
        a,  b : int
            Foo bar baz.
        """


class BadReturns:
    def return_not_documented(self):
        """
        Lacks section for Returns
        """
        return "Hello world!"

    def yield_not_documented(self):
        """
        Lacks section for Yields
        """
        yield "Hello world!"

    def no_type(self):
        """
        Returns documented but without type.

        Returns
        -------
        Some value.
        """
        return "Hello world!"

    def no_description(self):
        """
        Provides type but no description.

        Returns
        -------
        str
        """
        return "Hello world!"

    def no_punctuation(self):
        """
        Provides type and description but no period.

        Returns
        -------
        str
           A nice greeting
        """
        return "Hello world!"

    def named_single_return(self):
        """
        Provides name but returns only one value.

        Returns
        -------
        s : str
           A nice greeting.
        """
        return "Hello world!"

    def no_capitalization(self):
        """
        Forgets capitalization in return values description.

        Returns
        -------
        foo : str
           The first returned string.
        bar : str
           the second returned string.
        """
        return "Hello", "World!"

    def no_period_multi(self):
        """
        Forgets period in return values description.

        Returns
        -------
        foo : str
           The first returned string
        bar : str
           The second returned string.
        """
        return "Hello", "World!"


class BadSeeAlso:
    def no_desc(self):
        """
        Return the first 5 elements of the Series.

        See Also
        --------
        Series.tail
        """

    def desc_no_period(self):
        """
        Return the first 5 elements of the Series.

        See Also
        --------
        Series.tail : Return the last 5 elements of the Series.
        Series.iloc : Return a slice of the elements in the Series,
            which can also be used to return the first or last n
        """

    def desc_first_letter_lowercase(self):
        """
        Return the first 5 elements of the Series.

        See Also
        --------
        Series.tail : return the last 5 elements of the Series.
        Series.iloc : Return a slice of the elements in the Series,
            which can also be used to return the first or last n.
        """

    def prefix_pandas(self):
        """
        Have `pandas` prefix in See Also section.

        See Also
        --------
        pandas.Series.rename : Alter Series index labels or name.
        DataFrame.head : The first `n` rows of the caller object.
        """


class BadExamples:
    def missing_whitespace_around_arithmetic_operator(self):
        """
        Examples
        --------
        >>> 2 + 5
        7
        """

    def indentation_is_not_a_multiple_of_four(self):
        """
        Examples
        --------
        >>> if 2 + 5:
        ...     pass
        """

    def missing_whitespace_after_comma(self):
        """
        Examples
        --------
        >>> import datetime
        >>> value = datetime.date(2019, 1, 1)
        """


class ConstructorDocumentedInClassAndInit:
    """
    Class to test constructor documented via class and constructor docstrings.

    A case where both the class docstring and the constructor docstring are
    defined.

    Parameters
    ----------
    param1 : int
        Description of param1.

    See Also
    --------
    otherclass : A class that does something else.

    Examples
    --------
    This is an example of how to use ConstructorDocumentedInClassAndInit.
    """

    def __init__(self, param1: int) -> None:
        """
        Constructor docstring with additional information.

        Extended information.

        Parameters
        ----------
        param1 : int
            Description of param1 with extra details.

        See Also
        --------
        otherclass : A class that does something else.

        Examples
        --------
        This is an example of how to use ConstructorDocumentedInClassAndInit.
        """


class ConstructorDocumentedInClass:
    """
    Class to test constructor documented via class docstring.

    Useful to ensure that validation of `__init__` does not signal GL08,
    when the class docstring properly documents the `__init__` constructor.

    Parameters
    ----------
    param1 : int
        Description of param1.

    See Also
    --------
    otherclass : A class that does something else.

    Examples
    --------
    This is an example of how to use ConstructorDocumentedInClass.
    """

    def __init__(self, param1: int) -> None:
        pass


class ConstructorDocumentedInClassWithNoParameters:
    """
    Class to test constructor documented via class docstring with no parameters.

    Useful to ensure that validation of `__init__` does not signal GL08,
    when the class docstring properly documents the `__init__` constructor.

    See Also
    --------
    otherclass : A class that does something else.

    Examples
    --------
    This is an example of how to use ConstructorDocumentedInClassWithNoParameters.
    """

    def __init__(self) -> None:
        pass


class IncompleteConstructorDocumentedInClass:
    """
    Class to test an incomplete constructor docstring.

    This class does not properly document parameters.
    Unnecessary extended summary.

    See Also
    --------
    otherclass : A class that does something else.

    Examples
    --------
    This is an example of how to use IncompleteConstructorDocumentedInClass.
    """

    def __init__(self, param1: int):
        pass


class ConstructorDocumentedinEmbeddedClass:  # ignore Gl08, ES01
    """
    Class to test the initialisation behaviour of a embedded class.
    """

    class EmbeddedClass1:  # ignore GL08, ES01
        """
        An additional level for embedded class documentation checking.
        """

        class EmbeddedClass2:
            """
            This is an embedded class.

            Extended summary.

            Parameters
            ----------
            param1 : int
                Description of param1.

            See Also
            --------
            otherclass : A class that does something else.

            Examples
            --------
            This is an example of how to use EmbeddedClass.
            """

            def __init__(self, param1: int) -> None:
                pass


class IncompleteConstructorDocumentedinEmbeddedClass:
    """
    Class to test the initialisation behaviour of a embedded class.
    """

    class EmbeddedClass1:
        """
        An additional level for embedded class documentation checking.
        """

        class EmbeddedClass2:
            """
            This is an embedded class.

            Extended summary.

            See Also
            --------
            otherclass : A class that does something else.

            Examples
            --------
            This is an example of how to use EmbeddedClass.
            """

            def __init__(self, param1: int) -> None:
                pass


@dataclass
class DataclassWithDocstring:
    """
    A class decorated by `dataclass`.

    To check the functionality of `dataclass` objects do not break the Validator.
    As param1 is not documented this class should also raise PR01.
    """

    param1: int


class ClassWithPropertyObject:
    """
    A class with a `property`.

    To check the functionality of `property` objects do not break the Validator.

    Parameters
    ----------
    param1 : int
        Description of param1.
    """

    def __init__(self, param1: int) -> None:
        self._param1 = param1

    @property
    def param1(self) -> int:
        """
        Get the value of param1.

        Returns
        -------
        int
            The value of param1.
        """
        return self._param1


class TestValidator:
    def _import_path(self, klass=None, func=None):
        """
        Build the required import path for tests in this module.

        Parameters
        ----------
        klass : str
            Class name of object in module.
        func : str
            Function name of object in module.

        Returns
        -------
        str
            Import path of specified object in this module
        """
        base_path = "numpydoc.tests.test_validate"

        if klass:
            base_path = ".".join([base_path, klass])

        if func:
            base_path = ".".join([base_path, func])

        return base_path

    def test_one_liner(self, capsys):
        result = validate_one(
            self._import_path(klass="GoodDocStrings", func="one_liner")
        )
        errors = " ".join(err[1] for err in result["errors"])
        assert (
            "should start in the line immediately after the opening quotes"
            not in errors
        )
        assert "should be placed in the line after the last text" not in errors

    def test_good_class(self, capsys):
        errors = validate_one(self._import_path(klass="GoodDocStrings"))["errors"]
        assert isinstance(errors, list)
        assert not errors

    @pytest.mark.parametrize(
        "func",
        [
            "plot",
            "swap",
            "sample",
            "random_letters",
            "sample_values",
            "head",
            "head1",
            "summary_starts_with_number",
            "contains",
            "mode",
            "good_imports",
            "no_returns",
            "empty_returns",
            "multiple_variables_on_one_line",
            "other_parameters",
            "warnings",
            "valid_options_in_parameter_description_sets",
            "parameters_with_trailing_underscores",
            "parameter_with_wrong_types_as_substrings",
        ],
    )
    def test_good_functions(self, capsys, func):
        errors = validate_one(self._import_path(klass="GoodDocStrings", func=func))[
            "errors"
        ]
        assert isinstance(errors, list)
        assert not errors

    def test_bad_class(self, capsys):
        errors = validate_one(self._import_path(klass="BadGenericDocStrings"))["errors"]
        assert isinstance(errors, list)
        assert errors

    @pytest.mark.parametrize(
        "func",
        [
            "too_short_header_underline",
        ],
    )
    def test_bad_generic_functions(self, capsys, func):
        with pytest.warns(UserWarning, match="is too short"):
            validate_one(self._import_path(klass="WarnGenericFormat", func=func))

    @pytest.mark.parametrize(
        "func",
        [
            "func",
            "astype",
            "astype1",
            "astype2",
            "astype3",
            "plot",
            "directives_without_two_colons",
        ],
    )
    def test_bad_generic_functions(self, capsys, func):
        errors = validate_one(
            self._import_path(klass="BadGenericDocStrings", func=func)
        )["errors"]
        assert isinstance(errors, list)
        assert errors

    @pytest.mark.parametrize(
        "klass,func,msgs",
        [
            # See Also tests
            (
                "BadGenericDocStrings",
                "unknown_section",
                ('Found unknown section "Unknown Section".',),
            ),
            (
                "BadGenericDocStrings",
                "sections_in_wrong_order",
                (
                    "Sections are in the wrong order. Correct order is: Parameters, "
                    "See Also, Examples",
                ),
            ),
            (
                "BadGenericDocStrings",
                "deprecation_in_wrong_order",
                ("Deprecation warning should precede extended summary",),
            ),
            (
                "BadGenericDocStrings",
                "directives_without_two_colons",
                (
                    "reST directives ['versionchanged', 'versionadded', "
                    "'deprecated'] must be followed by two colons",
                ),
            ),
            (
                "BadSeeAlso",
                "no_desc",
                ('Missing description for See Also "Series.tail" reference',),
            ),
            (
                "BadSeeAlso",
                "desc_no_period",
                ('Missing period at end of description for See Also "Series.iloc"',),
            ),
            (
                "BadSeeAlso",
                "desc_first_letter_lowercase",
                ('should be capitalized for See Also "Series.tail"',),
            ),
            # Summary tests
            (
                "BadSummaries",
                "no_summary",
                ("No summary found",),
            ),
            (
                "BadSummaries",
                "heading_whitespaces",
                ("Summary contains heading whitespaces",),
            ),
            (
                "BadSummaries",
                "wrong_line",
                (
                    "should start in the line immediately after the opening quotes",
                    "should be placed in the line after the last text",
                ),
            ),
            ("BadSummaries", "no_punctuation", ("Summary does not end with a period",)),
            (
                "BadSummaries",
                "no_capitalization",
                ("Summary does not start with a capital letter",),
            ),
            (
                "BadSummaries",
                "no_capitalization",
                ("Summary must start with infinitive verb",),
            ),
            ("BadSummaries", "multi_line", ("Summary should fit in a single line",)),
            (
                "BadSummaries",
                "two_paragraph_multi_line",
                ("Summary should fit in a single line",),
            ),
            # Parameters tests
            (
                "BadParameters",
                "no_type",
                ('Parameter "value" has no type',),
            ),
            (
                "BadParameters",
                "type_with_period",
                ('Parameter "value" type should not finish with "."',),
            ),
            (
                "BadParameters",
                "no_description",
                ('Parameter "value" has no description',),
            ),
            (
                "BadParameters",
                "missing_params",
                ("Parameters {'**kwargs'} not documented",),
            ),
            (
                "BadParameters",
                "bad_colon_spacing",
                (
                    'Parameter "kind" requires a space before the colon '
                    "separating the parameter name and type",
                ),
            ),
            (
                "BadParameters",
                "no_description_period",
                ('Parameter "kind" description should finish with "."',),
            ),
            (
                "BadParameters",
                "no_description_period_with_directive",
                ('Parameter "kind" description should finish with "."',),
            ),
            (
                "BadParameters",
                "parameter_capitalization",
                ('Parameter "kind" description should start with a capital letter',),
            ),
            (
                "BadParameters",
                "integer_parameter",
                ('Parameter "kind" type should use "int" instead of "integer"',),
            ),
            (
                "BadParameters",
                "string_parameter",
                ('Parameter "kind" type should use "str" instead of "string"',),
            ),
            (
                "BadParameters",
                "boolean_parameter",
                ('Parameter "kind" type should use "bool" instead of "boolean"',),
            ),
            (
                "BadParameters",
                "list_incorrect_parameter_type",
                ('Parameter "kind" type should use "bool" instead of "boolean"',),
            ),
            (
                "BadParameters",
                "list_incorrect_parameter_type",
                ('Parameter "kind" type should use "int" instead of "integer"',),
            ),
            (
                "BadParameters",
                "list_incorrect_parameter_type",
                ('Parameter "kind" type should use "str" instead of "string"',),
            ),
            (
                "BadParameters",
                "bad_parameter_spacing",
                ("Parameters {'b'} not documented", "Unknown parameters {' b'}"),
            ),
            pytest.param(
                "BadParameters",
                "blank_lines",
                ("No error yet?",),
                marks=pytest.mark.xfail,
            ),
            # Returns tests
            ("BadReturns", "return_not_documented", ("No Returns section found",)),
            ("BadReturns", "yield_not_documented", ("No Yields section found",)),
            pytest.param("BadReturns", "no_type", ("foo",), marks=pytest.mark.xfail),
            ("BadReturns", "no_description", ("Return value has no description",)),
            (
                "BadReturns",
                "no_punctuation",
                ('Return value description should finish with "."',),
            ),
            (
                "BadReturns",
                "named_single_return",
                (
                    "The first line of the Returns section should contain only the "
                    "type, unless multiple values are being returned",
                ),
            ),
            (
                "BadReturns",
                "no_capitalization",
                ("Return value description should start with a capital letter",),
            ),
            (
                "BadReturns",
                "no_period_multi",
                ('Return value description should finish with "."',),
            ),
            (
                "BadGenericDocStrings",
                "method_wo_docstrings",
                ("The object does not have a docstring",),
            ),
            (
                "BadGenericDocStrings",
                "two_linebreaks_between_sections",
                (
                    "Double line break found; please use only one blank line to "
                    "separate sections or paragraphs, and do not leave blank lines "
                    "at the end of docstrings",
                ),
            ),
            (
                "BadGenericDocStrings",
                "linebreak_at_end_of_docstring",
                (
                    "Double line break found; please use only one blank line to "
                    "separate sections or paragraphs, and do not leave blank lines "
                    "at the end of docstrings",
                ),
            ),
        ],
    )
    def test_bad_docstrings(self, capsys, klass, func, msgs):
        with warnings.catch_warnings(record=True) as w:
            result = validate_one(self._import_path(klass=klass, func=func))
        if len(w):
            assert all("Unknown section" in str(ww.message) for ww in w)
        for msg in msgs:
            assert msg in " ".join(err[1] for err in result["errors"])

    @pytest.mark.parametrize(
        "klass,exp_init_codes,exc_init_codes,exp_klass_codes",
        [
            ("ConstructorDocumentedInClass", tuple(), ("GL08",), tuple()),
            ("ConstructorDocumentedInClassAndInit", tuple(), ("GL08",), tuple()),
            (
                "ConstructorDocumentedInClassWithNoParameters",
                tuple(),
                ("GL08",),
                tuple(),
            ),
            (
                "IncompleteConstructorDocumentedInClass",
                ("GL08",),
                tuple(),
                ("PR01"),  # Parameter not documented in class constructor
            ),
            (
                "ConstructorDocumentedinEmbeddedClass.EmbeddedClass1.EmbeddedClass2",
                tuple(),
                ("GL08",),
                tuple(),
            ),
            (
                "IncompleteConstructorDocumentedinEmbeddedClass.EmbeddedClass1.EmbeddedClass2",
                ("GL08",),
                tuple(),
                ("PR01",),
            ),
        ],
    )
    def test_constructor_docstrings(
        self, klass, exp_init_codes, exc_init_codes, exp_klass_codes
    ):
        # First test the class docstring itself, checking expected_klass_codes match
        result = validate_one(self._import_path(klass=klass))
        for err in result["errors"]:
            assert err[0] in exp_klass_codes

        # Then test the constructor docstring
        result = validate_one(self._import_path(klass=klass, func="__init__"))
        for code in exp_init_codes:
            assert code in " ".join(err[0] for err in result["errors"])
        for code in exc_init_codes:
            assert code not in " ".join(err[0] for err in result["errors"])

        if klass == "ConstructorDocumentedinEmbeddedClass":
            raise NotImplementedError(
                "Test for embedded class constructor docstring not implemented yet."
            )

    def test_dataclass_object(self):
        # Test validator methods complete execution on dataclass objects and methods
        # Test case ought to be removed if dataclass objects properly supported.
        result = validate_one(self._import_path(klass="DataclassWithDocstring"))
        # Check codes match as expected for dataclass objects.
        errs = ["ES01", "SA01", "EX01", "PR01"]
        for error in result["errors"]:
            assert error[0] in errs
            errs.remove(error[0])

        # Test initialisation method (usually undocumented in dataclass) raises any errors.
        init_fn = self._import_path(klass="DataclassWithDocstring", func="__init__")
        result = validate_one(init_fn)
        # Check that __init__ raises GL08 when the class docstring doesn't document params.
        assert result["errors"][0][0] == "GL08"

    def test_property_object(self):
        # Test validator methods complete execution on class property objects
        # Test case ought to be removed if property objects properly supported.
        result = validate_one(
            self._import_path(klass="ClassWithPropertyObject", func="param1")
        )
        # Check codes match as expected for property objects.
        errs = ["ES01", "SA01", "EX01"]
        for error in result["errors"]:
            assert error[0] in errs
            errs.remove(error[0])


def decorator(x):
    """Test decorator."""
    return x


@decorator
@decorator
class DecoratorClass:
    """
    Class and methods with decorators.

    * `DecoratorClass` has two decorators.
    * `DecoratorClass.test_no_decorator` has no decorator.
    * `DecoratorClass.test_property` has a `@property` decorator.
    * `DecoratorClass.test_cached_property` has a `@cached_property` decorator.
    * `DecoratorClass.test_three_decorators` has three decorators.

    `Validator.source_file_def_line` should return the `def` or `class` line number, not
    the line of the first decorator.
    """

    def test_no_decorator(self):
        """Test method without decorators."""

    @property
    def test_property(self):
        """Test property method."""

    @cached_property
    def test_cached_property(self):
        """Test property method."""

    @decorator
    @decorator
    @decorator
    def test_three_decorators(self):
        """Test method with three decorators."""


class TestValidatorClass:
    @pytest.mark.parametrize("invalid_name", ["unknown_mod", "unknown_mod.MyClass"])
    def test_raises_for_invalid_module_name(self, invalid_name):
        msg = f'No module can be imported from "{invalid_name}"'
        with pytest.raises(ImportError, match=msg):
            numpydoc.validate.Validator._load_obj(invalid_name)

    @pytest.mark.parametrize(
        "invalid_name", ["datetime.BadClassName", "datetime.bad_method_name"]
    )
    def test_raises_for_invalid_attribute_name(self, invalid_name):
        name_components = invalid_name.split(".")
        obj_name, invalid_attr_name = name_components[-2], name_components[-1]
        msg = f"'{obj_name}' has no attribute '{invalid_attr_name}'"
        with pytest.raises(AttributeError, match=msg):
            numpydoc.validate.Validator._load_obj(invalid_name)

    @pytest.mark.parametrize(
        ["decorated_obj", "def_line"],
        [
            [
                "numpydoc.tests.test_validate.DecoratorClass",
                getsourcelines(DecoratorClass)[-1] + 2,
            ],
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_no_decorator",
                getsourcelines(DecoratorClass.test_no_decorator)[-1],
            ],
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_property",
                getsourcelines(DecoratorClass.test_property.fget)[-1] + 1,
            ],
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_cached_property",
                getsourcelines(DecoratorClass.test_cached_property.func)[-1] + 1,
            ],
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_three_decorators",
                getsourcelines(DecoratorClass.test_three_decorators)[-1] + 3,
            ],
        ],
    )
    def test_source_file_def_line_with_decorators(self, decorated_obj, def_line):
        doc = numpydoc.validate.Validator(
            numpydoc.docscrape.get_doc_object(
                numpydoc.validate.Validator._load_obj(decorated_obj)
            )
        )
        assert doc.source_file_def_line == def_line

    @pytest.mark.parametrize(
        ["property", "file_name"],
        [
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_property",
                getsourcefile(DecoratorClass.test_property.fget),
            ],
            [
                "numpydoc.tests.test_validate.DecoratorClass.test_cached_property",
                getsourcefile(DecoratorClass.test_cached_property.func),
            ],
        ],
    )
    def test_source_file_name_with_properties(self, property, file_name):
        doc = numpydoc.validate.Validator(
            numpydoc.docscrape.get_doc_object(
                numpydoc.validate.Validator._load_obj(property)
            )
        )
        assert doc.source_file_name == file_name


def test_is_generator_validation_with_decorator():
    """Ensure that the check for a Yields section when an object is a generator
    (YD01) works with decorated generators."""

    def tinsel(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapper

    def foo():
        """A simple generator"""
        yield from range(10)

    @tinsel
    def bar():
        """Generator wrapped once"""
        yield from range(10)

    @tinsel
    @tinsel
    @tinsel
    def baz():
        """Generator wrapped multiple times"""
        yield from range(10)

    # foo without wrapper is a generator
    v = Validator(get_doc_object(foo))
    assert v.is_generator_function

    # Wrapped once
    v = Validator(get_doc_object(bar))
    assert v.is_generator_function

    # Wrapped multiple times
    v = Validator(get_doc_object(baz))
    assert v.is_generator_function



from collections import defaultdict
from copy import deepcopy
from io import StringIO
from pathlib import PosixPath

import pytest
from docutils import nodes
from sphinx.ext.autodoc import ALL
from sphinx.util import logging

from numpydoc.numpydoc import (
    _clean_text_signature,
    clean_backrefs,
    mangle_docstrings,
    update_config,
)
from numpydoc.xref import DEFAULT_LINKS


class MockConfig:
    numpydoc_use_plots = False
    numpydoc_show_class_members = True
    numpydoc_show_inherited_class_members = True
    numpydoc_class_members_toctree = True
    numpydoc_xref_param_type = False
    numpydoc_xref_aliases = {}
    numpydoc_xref_aliases_complete = deepcopy(DEFAULT_LINKS)
    numpydoc_xref_ignore = set()
    templates_path = []
    numpydoc_citation_re = "[a-z0-9_.-]+"
    numpydoc_attributes_as_param_list = True
    numpydoc_validation_checks = set()
    numpydoc_validation_exclude = set()
    numpydoc_validation_exclude_files = set()
    numpydoc_validation_overrides = dict()


class MockBuilder:
    config = MockConfig()
    _translator = None


class MockApp:
    config = MockConfig()
    builder = MockBuilder()
    translator = None

    def __init__(self):
        self.builder.app = self
        # Attrs required for logging
        self.verbosity = 2
        self._warncount = 0
        self.warningiserror = False
        self._exception_on_warning = False


def test_mangle_docstrings_basic():
    s = """
A top section before

.. autoclass:: str
    """
    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {}, lines)
    assert "rpartition" in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {"members": ["upper"]}, lines)
    assert "rpartition" not in [x.strip() for x in lines]
    assert "upper" in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(MockApp(), "class", "str", str, {"exclude-members": ALL}, lines)
    assert "rpartition" not in [x.strip() for x in lines]
    assert "upper" not in [x.strip() for x in lines]

    lines = s.split("\n")
    mangle_docstrings(
        MockApp(), "class", "str", str, {"exclude-members": ["upper"]}, lines
    )
    assert "rpartition" in [x.strip() for x in lines]
    assert "upper" not in [x.strip() for x in lines]


def test_mangle_docstrings_inherited_class_members():
    # if subclass docs are rendered, this PosixPath should have Path.samefile
    p = """
A top section before

.. autoclass:: pathlib.PosixPath
"""
    lines = p.split("\n")
    app = MockApp()
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" in lines
    app.config.numpydoc_show_inherited_class_members = False
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" not in lines
    app.config.numpydoc_show_inherited_class_members = dict()
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" in lines
    app.config.numpydoc_show_inherited_class_members = defaultdict(bool)
    lines = p.split("\n")
    mangle_docstrings(app, "class", "pathlib.PosixPath", PosixPath, {}, lines)
    lines = [x.strip() for x in lines]
    assert "samefile" not in lines


def test_clean_text_signature():
    assert _clean_text_signature(None) is None
    assert _clean_text_signature("func($self)") == "func()"
    assert (
        _clean_text_signature("func($self, *args, **kwargs)") == "func(*args, **kwargs)"
    )
    assert _clean_text_signature("($self)") == "()"
    assert _clean_text_signature("()") == "()"
    assert _clean_text_signature("func()") == "func()"
    assert (
        _clean_text_signature("func($self, /, *args, **kwargs)")
        == "func(*args, **kwargs)"
    )
    assert (
        _clean_text_signature("func($self, other, /, *args, **kwargs)")
        == "func(other, *args, **kwargs)"
    )
    assert _clean_text_signature("($module)") == "()"
    assert _clean_text_signature("func($type)") == "func()"
    assert (
        _clean_text_signature('func($self, foo="hello world")')
        == 'func(foo="hello world")'
    )
    assert (
        _clean_text_signature("func($self, foo='hello world')")
        == "func(foo='hello world')"
    )
    assert _clean_text_signature('func(foo="hello world")') == 'func(foo="hello world")'
    assert _clean_text_signature('func(foo="$self")') == 'func(foo="$self")'
    assert _clean_text_signature('func($self, foo="$self")') == 'func(foo="$self")'
    assert _clean_text_signature("func(self, other)") == "func(self, other)"
    assert _clean_text_signature("func($self, *args)") == "func(*args)"


@pytest.fixture
def f():
    def _function_without_seealso_and_examples():
        """
        A function whose docstring has no examples or see also section.

        Expect SA01 and EX01 errors if validation enabled.
        """

    return _function_without_seealso_and_examples


@pytest.mark.parametrize(
    (
        "numpydoc_validation_checks",
        "expected_warn",
        "non_warnings",
    ),
    (
        # Validation configured off - expect no warnings
        (set(), [], []),
        # Validation on with expected warnings
        ({"SA01", "EX01"}, ("SA01", "EX01"), []),
        # Validation on with only one activated check
        ({"SA01"}, ("SA01",), ("EX01",)),
    ),
)
def test_mangle_docstring_validation_warnings(
    f,
    numpydoc_validation_checks,
    expected_warn,
    non_warnings,
):
    app = MockApp()
    # Set up config for test
    app.config.numpydoc_validation_checks = numpydoc_validation_checks
    # Update configuration
    update_config(app)
    # Set up logging
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)
    # Run mangle docstrings with the above configuration
    mangle_docstrings(app, "function", "f", f, None, f.__doc__.split("\n"))
    # Assert that all (and only) expected warnings are logged
    warnings = warning.getvalue()
    for w in expected_warn:
        assert w in warnings
    for w in non_warnings:
        assert w not in warnings


def test_mangle_docstring_validation_exclude():
    def function_with_bad_docstring():
        """
        This docstring will raise docstring validation warnings."""

    app = MockApp()
    app.config.numpydoc_validation_checks = {"all"}
    app.config.numpydoc_validation_exclude = [r"_bad_"]
    # Call update_config to construct regexp from config value
    update_config(app)
    # Setup for catching warnings
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)
    # Run mangle docstrings on function_with_bad_docstring
    mangle_docstrings(
        app,
        "function",
        function_with_bad_docstring.__name__,
        function_with_bad_docstring,
        None,
        function_with_bad_docstring.__doc__.split("\n"),
    )
    # Validation is skipped due to exclude pattern matching fn name, therefore
    # no warnings expected
    assert warning.getvalue() == ""


@pytest.mark.parametrize("overrides", [{}, {"SS02"}, {"SS02", "SS03"}])
def test_mangle_docstrings_overrides(overrides):
    def process_something_noop_function():
        """Process something."""

    app = MockApp()
    app.config.numpydoc_validation_checks = {"all"}
    app.config.numpydoc_validation_overrides = {
        check: [r"^Process "]  # overrides are regex on docstring content
        for check in overrides
    }
    update_config(app)

    # Setup for catching warnings
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)

    # Run mangle docstrings on process_something_noop_function
    mangle_docstrings(
        app,
        "function",
        process_something_noop_function.__name__,
        process_something_noop_function,
        None,
        process_something_noop_function.__doc__.split("\n"),
    )

    findings = warning.getvalue()
    assert " EX01: " in findings  # should always be there
    for check in overrides:
        assert f" {check}: " not in findings


def test_update_config_invalid_validation_set():
    app = MockApp()
    # Results in {'a', 'l'} instead of {"all"}
    app.config.numpydoc_validation_checks = set("all")
    with pytest.raises(ValueError, match="Unrecognized validation code"):
        update_config(app)


def test_update_config_exclude_str():
    app = MockApp()
    app.config.numpydoc_validation_checks = set()
    app.config.numpydoc_validation_exclude = "shouldnt-be-a-str"
    with pytest.raises(ValueError, match=r"\['shouldnt-be-a-str'\]"):
        update_config(app)


def test_clean_backrefs():
    """Check ids are not cleaned from inline backrefs."""
    par = nodes.paragraph(rawsource="", text="")
    inline_ref = nodes.inline(rawsource="", text="", ids=["id1"])
    inline_ref += nodes.reference(rawsource="", text="[1]", refid="r123-1")
    citation = nodes.citation(
        rawsource="", docname="index", backrefs=["id1"], ids=["r123-1"]
    )
    citation += nodes.label("1")
    citation += nodes.paragraph(rawsource="", text="Author. Title.")
    par += inline_ref
    par += citation
    clean_backrefs(app=MockApp(), doc=par, docname="index")
    assert "id1" in citation["backrefs"]


@pytest.mark.parametrize(
    "exclude_files, has_warnings",
    [
        (
            [
                r"^doesnt_match_any_file$",
            ],
            True,
        ),
        (
            [
                r"^.*test_numpydoc\.py$",
            ],
            False,
        ),
    ],
)
def test_mangle_skip_exclude_files(exclude_files, has_warnings):
    """
    Check that the regex expressions in numpydoc_validation_files_exclude
    are correctly used to skip checks on files that match the patterns.
    """

    def process_something_noop_function():
        """Process something."""

    app = MockApp()
    app.config.numpydoc_validation_checks = {"all"}

    # Class attributes for config persist - need to reset them to unprocessed states.
    app.config.numpydoc_validation_exclude = set()  # Reset to default...
    app.config.numpydoc_validation_overrides = dict()  # Reset to default...

    app.config.numpydoc_validation_exclude_files = exclude_files
    update_config(app)

    # Setup for catching warnings
    status, warning = StringIO(), StringIO()
    logging.setup(app, status, warning)

    # Simulate a file that matches the exclude pattern
    mangle_docstrings(
        app,
        "function",
        process_something_noop_function.__name__,
        process_something_noop_function,
        None,
        process_something_noop_function.__doc__.split("\n"),
    )

    # Are warnings generated?
    print(warning.getvalue())
    assert bool(warning.getvalue()) is has_warnings


if __name__ == "__main__":
    import pytest

    pytest.main()



"""The CLI for numpydoc."""

import argparse
import ast
from collections.abc import Sequence
from pathlib import Path
from typing import List

from .docscrape_sphinx import get_doc_object
from .hooks import utils, validate_docstrings
from .validate import ERROR_MSGS, Validator, validate


def render_object(import_path: str, config: List[str] | None = None) -> int:
    """Test numpydoc docstring generation for a given object."""
    # TODO: Move Validator._load_obj to a better place than validate
    print(get_doc_object(Validator._load_obj(import_path), config=dict(config or [])))
    return 0


def validate_object(import_path: str) -> int:
    """Run numpydoc docstring validation for a given object."""
    exit_status = 0
    results = validate(import_path)
    for err_code, err_desc in results["errors"]:
        exit_status += 1
        print(":".join([import_path, err_code, err_desc]))
    return exit_status


def get_parser() -> argparse.ArgumentParser:
    """
    Build an argument parser.

    Returns
    -------
    argparse.ArgumentParser
        The argument parser.
    """
    ap = argparse.ArgumentParser(prog="numpydoc", description=__doc__)
    subparsers = ap.add_subparsers(title="subcommands")

    def _parse_config(s):
        key, _, value = s.partition("=")
        value = ast.literal_eval(value)
        return key, value

    render = subparsers.add_parser(
        "render",
        description="Generate an expanded RST-version of the docstring.",
        help="generate the RST docstring with numpydoc",
    )
    render.add_argument("import_path", help="e.g. numpy.ndarray")
    render.add_argument(
        "-c",
        "--config",
        type=_parse_config,
        action="append",
        help="key=val where val will be parsed by literal_eval, "
        "e.g. -c use_plots=True. Multiple -c can be used.",
    )
    render.set_defaults(func=render_object)

    validate = subparsers.add_parser(
        "validate",
        description="Validate an object's docstring against the numpydoc standard.",
        help="validate the object's docstring and report errors",
    )
    validate.add_argument("import_path", help="e.g. numpy.ndarray")
    validate.set_defaults(func=validate_object)

    project_root_from_cwd, config_file = utils.find_project_root(["."])
    config_options = validate_docstrings.parse_config(project_root_from_cwd)
    ignored_checks = [
        f"- {check}: {ERROR_MSGS[check]}"
        for check in set(ERROR_MSGS.keys()) - config_options["checks"]
    ]
    ignored_checks_text = "\n  " + "\n  ".join(ignored_checks) + "\n"

    lint_parser = subparsers.add_parser(
        "lint",
        description="Run numpydoc validation on files with option to ignore individual checks.",
        help="validate all docstrings in file(s) using the abstract syntax tree",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    lint_parser.add_argument(
        "files", type=str, nargs="+", help="File(s) to run numpydoc validation on."
    )
    lint_parser.add_argument(
        "--config",
        type=str,
        help=(
            "Path to a directory containing a pyproject.toml or setup.cfg file.\n"
            "The hook will look for it in the root project directory.\n"
            "If both are present, only pyproject.toml will be used.\n"
            "Options must be placed under\n"
            "    - [tool:numpydoc_validation] for setup.cfg files and\n"
            "    - [tool.numpydoc_validation] for pyproject.toml files."
        ),
    )
    lint_parser.add_argument(
        "--ignore",
        type=str,
        nargs="*",
        help=(
            f"""Check codes to ignore.{
                " Currently ignoring the following from "
                f"{Path(project_root_from_cwd) / config_file}: {ignored_checks_text}"
                "Values provided here will be in addition to the above, unless an alternate config is provided."
                if ignored_checks
                else ""
            }"""
        ),
    )
    lint_parser.set_defaults(func=validate_docstrings.run_hook)

    return ap


def main(argv: Sequence[str] | None = None) -> int:
    """CLI for numpydoc."""
    ap = get_parser()

    args = vars(ap.parse_args(argv))

    try:
        func = args.pop("func")
        return func(**args)
    except KeyError:
        ap.exit(status=2, message=ap.format_help())



"""Test the numpydoc validate pre-commit hook."""

import inspect
import re
from pathlib import Path

import pytest

from numpydoc.hooks.validate_docstrings import run_hook


@pytest.fixture
def example_module(request):
    fullpath = (
        Path(request.config.rootdir)
        / "numpydoc"
        / "tests"
        / "hooks"
        / "example_module.py"
    )
    return str(fullpath.relative_to(request.config.rootdir))


@pytest.mark.parametrize("config", [None, "fake_dir"])
def test_validate_hook(example_module, config, capsys):
    """Test that a file is correctly processed in the absence of config files."""

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: ES01 No extended summary found

        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:4: SA01 See Also section not found

        {example_module!s}:4: EX01 No examples section found

        {example_module!s}:8: ES01 No extended summary found

        {example_module!s}:8: SA01 See Also section not found

        {example_module!s}:8: EX01 No examples section found

        {example_module!s}:11: GL08 The object does not have a docstring

        {example_module!s}:17: ES01 No extended summary found

        {example_module!s}:17: PR01 Parameters {{'**kwargs'}} not documented

        {example_module!s}:17: PR07 Parameter "*args" has no description

        {example_module!s}:17: SA01 See Also section not found

        {example_module!s}:17: EX01 No examples section found

        {example_module!s}:26: SS05 Summary must start with infinitive verb, not third person (e.g. use "Generate" instead of "Generates")

        {example_module!s}:26: ES01 No extended summary found

        {example_module!s}:26: SA01 See Also section not found

        {example_module!s}:26: EX01 No examples section found

        {example_module!s}:30: GL08 The object does not have a docstring

        {example_module!s}:31: SA01 See Also section not found

        {example_module!s}:31: EX01 No examples section found

        {example_module!s}:46: SA01 See Also section not found

        {example_module!s}:46: EX01 No examples section found

        {example_module!s}:58: ES01 No extended summary found

        {example_module!s}:58: PR01 Parameters {{'name'}} not documented

        {example_module!s}:58: SA01 See Also section not found

        {example_module!s}:58: EX01 No examples section found
        """
    )

    return_code = run_hook([example_module], config=config)
    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


def test_validate_hook_with_ignore(example_module, capsys):
    """
    Test that a file is correctly processed in the absence of config files
    with command line ignore options.
    """

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:11: GL08 The object does not have a docstring

        {example_module!s}:17: PR01 Parameters {{'**kwargs'}} not documented

        {example_module!s}:17: PR07 Parameter "*args" has no description

        {example_module!s}:26: SS05 Summary must start with infinitive verb, not third person (e.g. use "Generate" instead of "Generates")

        {example_module!s}:30: GL08 The object does not have a docstring

        {example_module!s}:58: PR01 Parameters {{'name'}} not documented
        """
    )

    return_code = run_hook([example_module], ignore=["ES01", "SA01", "EX01"])

    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


def test_validate_hook_with_toml_config(example_module, tmp_path, capsys):
    """
    Test that a file is correctly processed with the config coming from
    a pyproject.toml file.
    """

    with open(tmp_path / "pyproject.toml", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                """
                [tool.numpydoc_validation]
                checks = [
                    "all",
                    "EX01",
                    "SA01",
                    "ES01",
                ]
                exclude = '\\.__init__$'
                override_SS05 = [
                    '^Creates',
                ]
                """
            )
        )

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:17: PR01 Parameters {{'**kwargs'}} not documented

        {example_module!s}:17: PR07 Parameter "*args" has no description

        {example_module!s}:30: GL08 The object does not have a docstring
        """
    )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


def test_validate_hook_with_setup_cfg(example_module, tmp_path, capsys):
    """
    Test that a file is correctly processed with the config coming from
    a setup.cfg file.
    """

    with open(tmp_path / "setup.cfg", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                """
                [tool:numpydoc_validation]
                checks = all,EX01,SA01,ES01
                exclude = \\.__init__$
                override_SS05 = ^Creates
                """
            )
        )

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:17: PR01 Parameters {{'**kwargs'}} not documented

        {example_module!s}:17: PR07 Parameter "*args" has no description

        {example_module!s}:30: GL08 The object does not have a docstring
        """
    )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


def test_validate_hook_exclude_option_pyproject(example_module, tmp_path, capsys):
    """
    Test that a file is correctly processed with the config coming from
    a pyproject.toml file and exclusions provided.
    """

    with open(tmp_path / "pyproject.toml", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                r"""
                [tool.numpydoc_validation]
                checks = [
                    "all",
                    "EX01",
                    "SA01",
                    "ES01",
                ]
                exclude = [
                    '\.do_something$',
                    '\.__init__$',
                ]
                override_SS05 = [
                    '^Creates',
                ]
                """
            )
        )

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:30: GL08 The object does not have a docstring
        """
    )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


def test_validate_hook_exclude_option_setup_cfg(example_module, tmp_path, capsys):
    """
    Test that a file is correctly processed with the config coming from
    a setup.cfg file and exclusions provided.
    """

    with open(tmp_path / "setup.cfg", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                """
                [tool:numpydoc_validation]
                checks = all,EX01,SA01,ES01
                exclude = \\.NewClass$,\\.__init__$
                override_SS05 = ^Creates
                """
            )
        )

    expected = inspect.cleandoc(
        f"""
        {example_module!s}:4: PR01 Parameters {{'name'}} not documented

        {example_module!s}:17: PR01 Parameters {{'**kwargs'}} not documented

        {example_module!s}:17: PR07 Parameter "*args" has no description
        """
    )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == 1
    assert capsys.readouterr().err.strip() == expected


@pytest.mark.parametrize(
    "file_exists, expected_code",
    [(True, 0), (False, 1)],
)
def test_validate_hook_exclude_files_option_pyproject(
    example_module, file_exists, expected_code, tmp_path
):
    """
    Test that the hook correctly processes the toml config and either includes
    or excludes files based on the `exclude_files` option.
    """
    exclude = str(example_module) if file_exists else "does_not_exist.py"

    with open(tmp_path / "pyproject.toml", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                f"""
                [tool.numpydoc_validation]
                checks = [
                    "all",
                    "EX01",
                    "SA01",
                    "ES01",
                ]
                exclude = '\\.__init__$'
                override_SS05 = [
                    '^Creates',
                ]
                exclude_files = [
                    '{re.escape(exclude)}',
                ]"""
            )
        )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == expected_code  # Should not-report/report findings.


@pytest.mark.parametrize(
    "file_exists, expected_code",
    [(True, 0), (False, 1)],
)
def test_validate_hook_exclude_files_option_setup_cfg(
    example_module, file_exists, expected_code, tmp_path
):
    """
    Test that the hook correctly processes the setup config and either includes
    or excludes files based on the `exclude_files` option.
    """
    exclude = str(example_module) if file_exists else "does_not_exist.py"

    with open(tmp_path / "setup.cfg", "w") as config_file:
        config_file.write(
            inspect.cleandoc(
                f"""
                [tool:numpydoc_validation]
                checks = all,EX01,SA01,ES01
                exclude = \\.NewClass$,\\.__init__$
                override_SS05 = ^Creates
                exclude_files = {re.escape(exclude)}
                """
            )
        )

    return_code = run_hook([example_module], config=tmp_path)
    assert return_code == expected_code  # Should not-report/report findings.
